import { ChoiceControl } from '../../lab/controls/ChoiceControl.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { JointUtil } from '../../lab/engine2D/JointUtil.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
import { Walls } from '../../lab/engine2D/Walls.js';
;
export class CarSuspensionApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-7, -5, 7, 5);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.springs = [];
        this.formation = 0;
        this.wheelMass = 0.1;
        this.carMass = 2;
        this.stiffness = 60;
        this.springDamping = 0.5;
        this.springLength = Math.sqrt(0.6 * 0.6 + 1);
        this.rbo.protoSpring.setWidth(0.3);
        this.protoWheel = new DisplayShape();
        this.protoWheel.setFillStyle('#CCF');
        this.sim.setShowForces(false);
        this.dampingLaw = new DampingLaw(0.1, 0.15, this.simList);
        this.elasticity.setElasticity(0.8);
        this.gravityLaw = new GravityLaw(10, this.simList);
        this.addPlaybackControls();
        let pn;
        this.addParameter(pn = new ParameterNumber(this, CarSuspensionApp.en.FORMATION, CarSuspensionApp.i18n.FORMATION, () => this.getFormation(), a => this.setFormation(a), [CarSuspensionApp.i18n.ROD_SPRING,
            CarSuspensionApp.i18n.TWO_SPRINGS], [0,
            1]));
        this.addControl(new ChoiceControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CarSuspensionApp.en.SPRING_DAMPING, CarSuspensionApp.i18n.SPRING_DAMPING, () => this.getSpringDamping(), a => this.setSpringDamping(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CarSuspensionApp.en.STIFFNESS, CarSuspensionApp.i18n.STIFFNESS, () => this.getStiffness(), a => this.setStiffness(a)));
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.configure();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'CarSuspensionApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('dampingLaw|gravityLaw|protoWheel', myName + '.');
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    addSpring(s) {
        s.setDamping(this.springDamping);
        this.sim.addForceLaw(s);
        this.sim.getSimList().add(s);
        this.springs.push(s);
    }
    ;
    configure() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        this.springs = [];
        const car = Shapes.makeBlock(5.0, 2.0, CarSuspensionApp.en.CAR, CarSuspensionApp.i18n.CAR);
        car.setMass(this.carMass);
        car.setPosition(new Vector(0, -1), 0);
        this.sim.addBody(car);
        this.displayList.findShape(car).setFillStyle('lightGray');
        switch (this.formation) {
            case 1:
                const wheel1 = Shapes.makeBall(0.4, CarSuspensionApp.en.WHEEL + 1, CarSuspensionApp.i18n.WHEEL + 1);
                wheel1.setMass(this.wheelMass);
                wheel1.setPosition(new Vector(-1.0, -3.0), 0);
                this.sim.addBody(wheel1);
                this.displayList.findShape(wheel1).setPrototype(this.protoWheel);
                const wheel2 = Shapes.makeBall(0.4, CarSuspensionApp.en.WHEEL + 2, CarSuspensionApp.i18n.WHEEL + 2);
                wheel2.setMass(this.wheelMass);
                wheel2.setPosition(new Vector(2.0, -3.0), 0);
                this.sim.addBody(wheel2);
                this.displayList.findShape(wheel2).setPrototype(this.protoWheel);
                this.addSpring(new Spring('spring1', car, new Vector(-2.1, -1.0), wheel1, Vector.ORIGIN, this.springLength, this.stiffness));
                this.addSpring(new Spring('spring2', car, new Vector(-0.9, -1.0), wheel1, Vector.ORIGIN, this.springLength, this.stiffness));
                this.addSpring(new Spring('spring3', car, new Vector(0.9, -1.0), wheel2, Vector.ORIGIN, this.springLength, this.stiffness));
                this.addSpring(new Spring('spring4', car, new Vector(2.1, -1.0), wheel2, Vector.ORIGIN, this.springLength, this.stiffness));
                break;
            case 0:
                const p1 = Shapes.makePendulum(0.05, this.springLength, 0.4, CarSuspensionApp.en.WHEEL + 1, CarSuspensionApp.i18n.WHEEL + 1);
                p1.setAngle(0);
                this.sim.addBody(p1);
                this.displayList.findShape(p1).setPrototype(this.protoWheel);
                JointUtil.attachRigidBody(this.sim, car, new Vector(-0.9, -1.0), p1, new Vector(0, this.springLength), 1);
                this.addSpring(new Spring('spring1', car, new Vector(-2.1, -1.0), p1, Vector.ORIGIN, this.springLength, this.stiffness));
                const p2 = Shapes.makePendulum(0.05, this.springLength, 0.4, CarSuspensionApp.en.WHEEL + 2, CarSuspensionApp.i18n.WHEEL + 2);
                p2.setAngle(0);
                this.sim.addBody(p2);
                this.displayList.findShape(p2).setPrototype(this.protoWheel);
                JointUtil.attachRigidBody(this.sim, car, new Vector(0.9, -1.0), p2, new Vector(0, this.springLength), 1);
                this.addSpring(new Spring('spring2', car, new Vector(2.1, -1.0), p2, Vector.ORIGIN, this.springLength, this.stiffness));
                this.sim.alignConnectors();
                break;
            default:
                throw '';
        }
        Walls.make(this.sim, 14, 10, 1.0);
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        this.sim.setElasticity(elasticity);
        this.sim.saveInitialState();
        this.clock.setTime(this.sim.getTime());
        this.clock.setRealTime(this.sim.getTime());
        this.easyScript.update();
    }
    ;
    getFormation() {
        return this.formation;
    }
    ;
    setFormation(value) {
        this.formation = value;
        this.configure();
        this.broadcastParameter(CarSuspensionApp.en.FORMATION);
    }
    ;
    getSpringDamping() {
        return this.springDamping;
    }
    ;
    setSpringDamping(value) {
        this.springDamping = value;
        for (let i = 0; i < this.springs.length; i++) {
            this.springs[i].setDamping(this.springDamping);
        }
        this.broadcastParameter(CarSuspensionApp.en.SPRING_DAMPING);
    }
    ;
    getStiffness() {
        return this.stiffness;
    }
    ;
    setStiffness(value) {
        this.stiffness = value;
        for (let i = 0; i < this.springs.length; i++) {
            this.springs[i].setStiffness(this.stiffness);
        }
        this.broadcastParameter(CarSuspensionApp.en.STIFFNESS);
    }
    ;
}
CarSuspensionApp.en = {
    CAR_MASS: 'car mass',
    FORMATION: 'formation',
    LENGTH: 'spring length',
    ROD_SPRING: 'rod and spring',
    SPRING_DAMPING: 'spring damping',
    STIFFNESS: 'spring stiffness',
    TWO_SPRINGS: 'two springs',
    WHEEL_MASS: 'wheel mass',
    WHEEL: 'wheel',
    CAR: 'car'
};
CarSuspensionApp.de_strings = {
    CAR_MASS: 'Auto Masse',
    FORMATION: 'Formation',
    LENGTH: 'Federlänge',
    ROD_SPRING: 'Stange und Feder',
    SPRING_DAMPING: 'Federdämpfung',
    STIFFNESS: 'Federsteifheit',
    TWO_SPRINGS: 'zwei Feder',
    WHEEL_MASS: 'Rad Masse',
    WHEEL: 'Rad',
    CAR: 'Auto'
};
CarSuspensionApp.i18n = Util.LOCALE === 'de' ? CarSuspensionApp.de_strings : CarSuspensionApp.en;
Util.defineGlobal('sims$engine2D$CarSuspensionApp', CarSuspensionApp);
