import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { SixThrusters } from './SixThrusters.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
import { Walls } from '../../lab/engine2D/Walls.js';
export class CurvedTestApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-4, -6, 8, 6);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.numBods = 6;
        this.thrust = 1.5;
        this.rbo.protoPolygon.setDrawCenterOfMass(true);
        this.rbo.protoPolygon.setNameFont('10pt sans-serif');
        this.rbo.protoSpring.setWidth(0.3);
        this.sim.setShowForces(true);
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(3.0, this.simList);
        this.elasticity.setElasticity(0.8);
        this.addPlaybackControls();
        let pn = new ParameterNumber(this, CurvedTestApp.en.NUM_BODIES, CurvedTestApp.i18n.NUM_BODIES, () => this.getNumBodies(), a => this.setNumBodies(a));
        pn.setDecimalPlaces(0);
        pn.setLowerLimit(1);
        pn.setUpperLimit(6);
        this.addParameter(pn);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CurvedTestApp.en.THRUST, CurvedTestApp.i18n.THRUST, () => this.getThrust(), a => this.setThrust(a)));
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'CurvedTestApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    config() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        CurvedTestApp.make(this.sim, this.gravityLaw, this.dampingLaw, this.numBods, this.simView.getSimRect(), this.displayList);
        let b;
        if (this.numBods >= 1) {
            b = this.sim.getBody('block1');
            this.thrust2 = SixThrusters.make(this.thrust, b);
            this.rbeh.setThrusters(this.thrust2, 'left');
            this.sim.addForceLaw(this.thrust2);
        }
        if (this.numBods >= 2) {
            b = this.sim.getBody('ball2');
            this.thrust1 = SixThrusters.make(this.thrust, b);
            this.rbeh.setThrusters(this.thrust1, 'right');
            this.sim.addForceLaw(this.thrust1);
        }
        this.sim.setElasticity(elasticity);
        this.sim.saveInitialState();
        this.clock.setTime(this.sim.getTime());
        this.clock.setRealTime(this.sim.getTime());
        this.easyScript.update();
    }
    ;
    static make(sim, gravity, damping, numBods, simRect, displayList) {
        sim.addForceLaw(damping);
        damping.connect(sim.getSimList());
        sim.addForceLaw(gravity);
        gravity.connect(sim.getSimList());
        let b;
        b = Shapes.makeBlock(2.0, 2.0, CurvedTestApp.en.FIX_BLOCK + 1, CurvedTestApp.i18n.FIX_BLOCK + 1);
        b.setMass(Infinity);
        b.setPosition(new Vector(-0.4, -4.6), -Math.PI / 16);
        sim.addBody(b);
        b = Shapes.makeBall(2, CurvedTestApp.en.FIX_BALL + 2, CurvedTestApp.i18n.FIX_BALL + 2);
        b.setMass(Infinity);
        b.setPosition(new Vector(3, 2.5 - 4.5 - 0.1), 0);
        sim.addBody(b);
        b = Shapes.makeBlock(2.0, 2.0, CurvedTestApp.en.FIX_BLOCK + 3, CurvedTestApp.i18n.FIX_BLOCK + 3);
        b.setMass(Infinity);
        b.setPosition(new Vector(-3.4, -4.6), Math.PI / 16);
        sim.addBody(b);
        b = Shapes.makeBlock(1.5, 4, CurvedTestApp.en.FIX_BLOCK + 4, CurvedTestApp.i18n.FIX_BLOCK + 4);
        b.setMass(Infinity);
        b.setPosition(new Vector(6.7, -1.0), -0.2);
        sim.addBody(b);
        if (numBods >= 1) {
            b = Shapes.makeBlock(0.8, 1.5, CurvedTestApp.en.BLOCK + 1, CurvedTestApp.i18n.BLOCK + 1);
            b.setPosition(new Vector(-2.0, -2), 0);
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('cyan');
            }
            ;
        }
        if (numBods >= 2) {
            b = Shapes.makeBall(0.8, CurvedTestApp.en.BALL + 2, CurvedTestApp.i18n.BALL + 2);
            b.setCenterOfMass(new Vector(b.getLeftBody() + 0.5 * b.getWidth(), b.getBottomBody() + 0.2 * b.getHeight()));
            b.setPosition(new Vector(-1.7, 1), 0);
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('orange');
            }
            ;
        }
        if (numBods >= 3) {
            b = Shapes.makeBall(1, CurvedTestApp.en.BALL + 2, CurvedTestApp.i18n.BALL + 2);
            let x = 0;
            let y = -1 + 2.0;
            x += 1 - 0.5;
            y += 1 - 0.3;
            b.setPosition(new Vector(x, 0.1 + y + sim.getDistanceTol() / 2.0), Math.PI);
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('#9f3');
            }
            ;
        }
        if (numBods >= 4) {
            b = Shapes.makeBlock(0.8, 2, CurvedTestApp.en.BLOCK + 4, CurvedTestApp.i18n.BLOCK + 4);
            b.setMass(2);
            b.setPosition(new Vector(5, 0), 0.2);
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('#f6c');
            }
            ;
        }
        if (numBods >= 5) {
            b = Shapes.makeBall(0.4, CurvedTestApp.en.BALL + 5, CurvedTestApp.i18n.BALL + 5);
            b.setMass(1.5);
            b.setPosition(new Vector(5.9, 2));
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('#39f');
            }
            ;
        }
        if (numBods >= 6) {
            b = Shapes.makeBall(1.0, CurvedTestApp.en.BALL + 6, CurvedTestApp.i18n.BALL + 6);
            b.setMass(1.0);
            b.setPosition(new Vector(-2.5, 4));
            sim.addBody(b);
            if (displayList != null) {
                displayList.findShape(b).setFillStyle('#c99');
            }
            ;
        }
        const zel = Walls.make2(sim, simRect);
        gravity.setZeroEnergyLevel(zel);
    }
    ;
    getNumBodies() {
        return this.numBods;
    }
    ;
    setNumBodies(value) {
        this.numBods = value;
        this.config();
        this.broadcastParameter(CurvedTestApp.en.NUM_BODIES);
    }
    ;
    getThrust() {
        return this.thrust;
    }
    ;
    setThrust(value) {
        this.thrust = value;
        this.thrust1.setMagnitude(value);
        this.thrust2.setMagnitude(value);
        this.broadcastParameter(CurvedTestApp.en.THRUST);
    }
    ;
}
CurvedTestApp.en = {
    NUM_BODIES: 'number of objects',
    THRUST: 'thrust',
    BALL: 'ball',
    BLOCK: 'block',
    FIX_BALL: 'fixed ball',
    FIX_BLOCK: 'fixed block'
};
CurvedTestApp.de_strings = {
    NUM_BODIES: 'Anzahl von Objekten',
    THRUST: 'Schubkraft',
    BALL: 'Ball',
    BLOCK: 'Block',
    FIX_BALL: 'Festball',
    FIX_BLOCK: 'Festblock'
};
CurvedTestApp.i18n = Util.LOCALE === 'de' ? CurvedTestApp.de_strings : CurvedTestApp.en;
Util.defineGlobal('sims$engine2D$CurvedTestApp', CurvedTestApp);
