import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { CommonControls } from '../common/CommonControls.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { Gravity2Law } from '../../lab/model/Gravity2Law.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class MarsMoonApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-10, -10, 10, 10);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.moonMass = 100;
        this.asteroidMass = 0.1;
        this.moonRadius = 3;
        this.asteroidRadius = 0.2;
        this.velocity = 14;
        this.distance = 0.8;
        this.ovalness = 1.0;
        this.moonSpin = 0;
        this.layout.getSimCanvas().setBackground('black');
        this.layout.getSimCanvas().setAlpha(CommonControls.SHORT_TRAILS);
        this.elasticity.setElasticity(1.0);
        this.sim.setShowForces(false);
        this.sim.setShowCollisions(true);
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new Gravity2Law(10.0, this.simList);
        this.addPlaybackControls();
        let pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(Gravity2Law.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.MOON_MASS, MarsMoonApp.i18n.MOON_MASS, () => this.getMoonMass(), a => this.setMoonMass(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.ASTERIOD_MASS, MarsMoonApp.i18n.ASTERIOD_MASS, () => this.getAsteroidMass(), a => this.setAsteroidMass(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.VELOCITY, MarsMoonApp.i18n.VELOCITY, () => this.getVelocity(), a => this.setVelocity(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.DISTANCE, MarsMoonApp.i18n.DISTANCE, () => this.getDistance(), a => this.setDistance(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.MOON_RADIUS, MarsMoonApp.i18n.MOON_RADIUS, () => this.getMoonRadius(), a => this.setMoonRadius(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MarsMoonApp.en.ASTERIOD_RADIUS, MarsMoonApp.i18n.ASTERIOD_RADIUS, () => this.getAsteroidRadius(), a => this.setAsteroidRadius(a)));
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
        this.panZoomParam.setValue(true);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'MarsMoonApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
        this.terminal.addRegex('MarsMoonApp|Engine2DApp', 'sims$engine2D$', false);
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    graphSetup(_body) {
        const idx = this.asteroid.getVarsIndex();
        this.graph.line.setXVariable(idx + 0);
        this.graph.line.setYVariable(idx + 2);
        this.timeGraph.line1.setYVariable(idx + 0);
        this.timeGraph.line2.setYVariable(idx + 1);
    }
    ;
    config() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        if (this.ovalness == 1.0) {
            this.moon = Shapes.makeBall(this.moonRadius, MarsMoonApp.en.MOON, MarsMoonApp.i18n.MOON);
        }
        else {
            throw 'oval shape not implemented';
        }
        const moon_v = -this.velocity * this.asteroidMass / this.moonMass;
        this.moon.setMass(this.moonMass);
        this.moon.setPosition(new Vector(0, 0), 0);
        this.moon.setVelocity(new Vector(0, moon_v), this.moonSpin);
        this.moon.setElasticity(elasticity);
        this.sim.addBody(this.moon);
        let ds = this.displayList.findShape(this.moon);
        ds.setFillStyle('#E0E0E0');
        ds.setDrawCenterOfMass(true);
        this.asteroid = Shapes.makeBall(this.asteroidRadius, MarsMoonApp.en.ASTERIOD, MarsMoonApp.i18n.ASTERIOD);
        const dist = this.asteroidRadius + this.moonRadius + this.distance;
        this.asteroid.setPosition(new Vector(dist, 0), 0);
        this.asteroid.setMass(this.asteroidMass);
        this.asteroid.setVelocity(new Vector(0, this.velocity), 0);
        this.asteroid.setElasticity(elasticity);
        this.sim.addBody(this.asteroid);
        this.displayList.findShape(this.asteroid).setFillStyle('Blue');
        this.sim.getVarsList().setTime(0);
        this.sim.saveInitialState();
        this.clock.setTime(this.sim.getTime());
        this.clock.setRealTime(this.sim.getTime());
        this.easyScript.update();
    }
    ;
    getAsteroidMass() {
        return this.asteroidMass;
    }
    ;
    setAsteroidMass(value) {
        this.asteroidMass = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.ASTERIOD_MASS);
    }
    ;
    getMoonMass() {
        return this.moonMass;
    }
    ;
    setMoonMass(value) {
        this.moonMass = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.MOON_MASS);
    }
    ;
    getAsteroidRadius() {
        return this.asteroidRadius;
    }
    ;
    setAsteroidRadius(value) {
        this.asteroidRadius = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.ASTERIOD_RADIUS);
    }
    ;
    getMoonRadius() {
        return this.moonRadius;
    }
    ;
    setMoonRadius(value) {
        this.moonRadius = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.MOON_RADIUS);
    }
    ;
    getVelocity() {
        return this.velocity;
    }
    ;
    setVelocity(value) {
        this.velocity = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.VELOCITY);
    }
    ;
    getDistance() {
        return this.distance;
    }
    ;
    setDistance(value) {
        this.distance = value;
        this.config();
        this.broadcastParameter(MarsMoonApp.en.DISTANCE);
    }
    ;
}
MarsMoonApp.en = {
    ASTERIOD_MASS: 'asteroid mass',
    ASTERIOD_RADIUS: 'asteroid radius',
    GRAVITY: 'gravity',
    MOON_MASS: 'moon mass',
    MOON_RADIUS: 'moon radius',
    MOON_SPIN: 'moon spin',
    OVALNESS: 'ovalness',
    VELOCITY: 'initial velocity',
    MOON: 'moon',
    ASTERIOD: 'asteroid',
    DISTANCE: 'distance'
};
MarsMoonApp.de_strings = {
    ASTERIOD_MASS: 'Asteroid Masse',
    ASTERIOD_RADIUS: 'Asteroid Radius',
    GRAVITY: 'Gravitation',
    MOON_MASS: 'Mondmasse',
    MOON_RADIUS: 'Mond Radius',
    MOON_SPIN: 'Mondrotation',
    OVALNESS: 'Oval-heit',
    VELOCITY: 'Anfangs Geschwindigkeit',
    MOON: 'Mond',
    ASTERIOD: 'Asteroid',
    DISTANCE: 'Entfernung'
};
MarsMoonApp.i18n = Util.LOCALE === 'de' ? MarsMoonApp.de_strings : MarsMoonApp.en;
Util.defineGlobal('sims$engine2D$MarsMoonApp', MarsMoonApp);
