import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { CommonControls } from '../common/CommonControls.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { JointUtil } from '../../lab/engine2D/JointUtil.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Scrim } from '../../lab/engine2D/Scrim.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class NewtonsCradleApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-5, -3, 5, 5);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.layout.getSimCanvas().setBackground('black');
        this.layout.getSimCanvas().setAlpha(CommonControls.SHORT_TRAILS);
        this.elasticity.setElasticity(1.0);
        this.sim.setShowForces(false);
        advance.setJointSmallImpacts(false);
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(32, this.simList);
        this.options = {
            stickLength: 3.0,
            radius: 0.6,
            gapDistance: 0,
            numBods: 5,
            startAngle: -Math.PI / 4
        };
        this.addPlaybackControls();
        let pn = new ParameterNumber(this, NewtonsCradleApp.en.NUM_BODIES, NewtonsCradleApp.i18n.NUM_BODIES, () => this.getNumBodies(), a => this.setNumBodies(a));
        pn.setDecimalPlaces(0);
        pn.setLowerLimit(1);
        pn.setUpperLimit(6);
        this.addParameter(pn);
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, NewtonsCradleApp.en.LENGTH, NewtonsCradleApp.i18n.LENGTH, () => this.getStickLength(), a => this.setStickLength(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, NewtonsCradleApp.en.GAP_DISTANCE, NewtonsCradleApp.i18n.GAP_DISTANCE, () => this.getGapDistance(), a => this.setGapDistance(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, NewtonsCradleApp.en.RADIUS, NewtonsCradleApp.i18n.RADIUS, () => this.getRadius(), a => this.setRadius(a)));
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'NewtonsCradleApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    graphSetup(body) {
        body = body ?? this.sim.getBody('pendulum1');
        this.graph.line.setXVariable(body.getVarsIndex() + 4);
        this.graph.line.setYVariable(body.getVarsIndex() + 5);
        this.timeGraph.line1.setYVariable(body.getVarsIndex() + 4);
        this.timeGraph.line2.setYVariable(body.getVarsIndex() + 5);
    }
    ;
    getNumBodies() {
        return this.options.numBods;
    }
    ;
    setNumBodies(value) {
        this.options.numBods = value;
        this.config();
        this.broadcastParameter(NewtonsCradleApp.en.NUM_BODIES);
    }
    ;
    getStickLength() {
        return this.options.stickLength;
    }
    ;
    setStickLength(value) {
        this.options.stickLength = value;
        this.config();
        this.broadcastParameter(NewtonsCradleApp.en.LENGTH);
    }
    ;
    getGapDistance() {
        return this.options.gapDistance;
    }
    ;
    setGapDistance(value) {
        this.options.gapDistance = value;
        this.config();
        this.broadcastParameter(NewtonsCradleApp.en.GAP_DISTANCE);
    }
    ;
    getRadius() {
        return this.options.radius;
    }
    ;
    setRadius(value) {
        this.options.radius = value;
        this.config();
        this.broadcastParameter(NewtonsCradleApp.en.RADIUS);
    }
    ;
    config() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        NewtonsCradleApp.make(this.sim, this.options);
        this.sim.setElasticity(elasticity);
        this.sim.getVarsList().setTime(0);
        this.sim.saveInitialState();
        this.clock.setTime(0);
        this.clock.setRealTime(0);
        this.easyScript.update();
    }
    ;
    static make(sim, options) {
        const n = options.numBods;
        const r = options.radius;
        const between = sim.getDistanceTol() / 2 + options.gapDistance;
        const tw = n * 2 * r + (n - 1) * between;
        let x = -tw / 2 + r;
        for (let i = 0; i < options.numBods; i++) {
            const body = Shapes.makePendulum(0.05, options.stickLength, options.radius, NewtonsCradleApp.en.PENDULUM + (i + 1), NewtonsCradleApp.i18n.PENDULUM + (i + 1));
            body.setAngle(0);
            sim.addBody(body);
            JointUtil.attachRigidBody(sim, Scrim.getScrim(), new Vector(x, 2), body, new Vector(0, options.stickLength), 1);
            body.setZeroEnergyLevel();
            sim.alignConnectors();
            x += body.getWidth() + between;
        }
        sim.getBody('pendulum1').setAngle(options.startAngle);
        sim.alignConnectors();
    }
    ;
}
NewtonsCradleApp.en = {
    GAP_DISTANCE: 'offset',
    PENDULUM: 'pendulum',
    NUM_BODIES: 'pendulums',
    LENGTH: 'pendulum length',
    RADIUS: 'radius'
};
NewtonsCradleApp.de_strings = {
    GAP_DISTANCE: 'Zwischenraum',
    PENDULUM: 'Pendel',
    NUM_BODIES: 'Pendelanzahl',
    LENGTH: 'Pendellänge',
    RADIUS: 'Radius'
};
NewtonsCradleApp.i18n = Util.LOCALE === 'de' ? NewtonsCradleApp.de_strings : NewtonsCradleApp.en;
Util.defineGlobal('sims$engine2D$NewtonsCradleApp', NewtonsCradleApp);
