import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ConstantForceLaw } from '../../lab/model/ConstantForceLaw.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { Force } from '../../lab/model/Force.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { PendulumClockConfig } from './PendulumClockConfig.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class PendulumClockApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-4, -4, 4, 6);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.pendulumLength = 3.0;
        this.turningForce = 1.0;
        this.extraBody = false;
        this.withGears = false;
        this.turnForceLaw = null;
        this.sim.setShowForces(true);
        this.dampingLaw = new DampingLaw(0.05, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(10, this.simList);
        this.elasticity.setElasticity(0.3);
        this.addPlaybackControls();
        let pb;
        let pn;
        this.addParameter(pb = new ParameterBoolean(this, PendulumClockConfig.en.WITH_GEARS, PendulumClockConfig.i18n.WITH_GEARS, () => this.getWithGears(), a => this.setWithGears(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pb = new ParameterBoolean(this, PendulumClockConfig.en.EXTRA_BODY, PendulumClockConfig.i18n.EXTRA_BODY, () => this.getExtraBody(), a => this.setExtraBody(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pn = new ParameterNumber(this, PendulumClockConfig.en.PENDULUM_LENGTH, PendulumClockConfig.i18n.PENDULUM_LENGTH, () => this.getPendulumLength(), a => this.setPendulumLength(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, PendulumClockConfig.en.TURNING_FORCE, PendulumClockConfig.i18n.TURNING_FORCE, () => this.getTurningForce(), a => this.setTurningForce(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'PendulumClockApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
        this.terminal.addRegex('PendulumClockApp|Engine2DApp', 'sims$engine2D$', false);
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    graphSetup(body) {
        body = body ?? this.sim.getBody(PendulumClockConfig.en.ANCHOR);
        this.graph.line.setXVariable(body.getVarsIndex() + 4);
        this.graph.line.setYVariable(body.getVarsIndex() + 5);
        this.timeGraph.line1.setYVariable(body.getVarsIndex() + 4);
        this.timeGraph.line2.setYVariable(body.getVarsIndex() + 5);
    }
    ;
    config() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        if (this.withGears) {
            PendulumClockConfig.makeClockWithGears(this.sim, this.pendulumLength, Vector.ORIGIN);
        }
        else {
            PendulumClockConfig.makeClock(this.sim, this.pendulumLength, Vector.ORIGIN);
        }
        const escapeWheel = this.displayList.findShape(PendulumClockConfig.en.ESCAPE_WHEEL);
        escapeWheel.setFillStyle('#D3D3D3');
        if (this.withGears) {
            escapeWheel.setStrokeStyle('black');
            const gear2 = this.displayList.findShape(PendulumClockConfig.en.GEAR + 2);
            gear2.setFillStyle('#B0C4DE');
            gear2.setZIndex(2);
        }
        const anchor = this.displayList.findShape(PendulumClockConfig.en.ANCHOR);
        anchor.setFillStyle('#B0C4DE');
        anchor.setZIndex(3);
        this.setTurningForce(this.turningForce);
        if (this.extraBody) {
            const block = Shapes.makeBlock(0.12, 0.5, PendulumClockConfig.en.EXTRA_BODY, PendulumClockConfig.i18n.EXTRA_BODY);
            block.setMass(0.3);
            block.setPosition(new Vector(0, 6), Math.PI / 2);
            this.sim.addBody(block);
            this.displayList.findShape(block).setStrokeStyle('cyan');
        }
        this.sim.setElasticity(elasticity);
        this.sim.getVarsList().setTime(0);
        this.sim.saveInitialState();
        this.clock.setTime(0);
        this.clock.setRealTime(0);
        this.easyScript.update();
    }
    ;
    getWithGears() {
        return this.withGears;
    }
    ;
    setWithGears(value) {
        this.withGears = value;
        this.config();
        this.broadcastParameter(PendulumClockConfig.en.WITH_GEARS);
    }
    ;
    getExtraBody() {
        return this.extraBody;
    }
    ;
    setExtraBody(value) {
        this.extraBody = value;
        this.config();
        this.broadcastParameter(PendulumClockConfig.en.EXTRA_BODY);
    }
    ;
    getPendulumLength() {
        return this.pendulumLength;
    }
    ;
    setPendulumLength(value) {
        this.pendulumLength = value;
        this.config();
        this.broadcastParameter(PendulumClockConfig.en.PENDULUM_LENGTH);
    }
    ;
    getTurningForce() {
        return this.turningForce;
    }
    ;
    setTurningForce(value) {
        this.turningForce = value;
        if (this.turnForceLaw != null) {
            this.sim.removeForceLaw(this.turnForceLaw);
        }
        const body = this.sim.getBody(PendulumClockConfig.en.ESCAPE_WHEEL);
        if (body) {
            const f = new Force('turning', body, body.getDragPoints()[0], 0, new Vector(value, 0), 0);
            this.turnForceLaw = new ConstantForceLaw(f);
            this.sim.addForceLaw(this.turnForceLaw);
        }
        this.broadcastParameter(PendulumClockConfig.en.TURNING_FORCE);
    }
    ;
}
Util.defineGlobal('sims$engine2D$PendulumClockApp', PendulumClockApp);
