import { AbstractSubject } from '../../lab/util/AbstractSubject.js';
import { DisplayAxes } from '../../lab/graph/DisplayAxes.js';
import { DisplayGraph } from '../../lab/graph/DisplayGraph.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { EasyScriptParser } from '../../lab/util/EasyScriptParser.js';
import { GenericObserver, ParameterNumber } from '../../lab/util/Observe.js';
import { GraphLine } from '../../lab/graph/GraphLine.js';
import { LabCanvas } from '../../lab/view/LabCanvas.js';
import { SimController } from '../../lab/app/SimController.js';
import { SimView } from '../../lab/view/SimView.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { Terminal } from '../../lab/util/Terminal.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
export class GraphCalc2App extends AbstractSubject {
    constructor(elem_ids, opt_name) {
        super(opt_name || 'GraphCalc2App');
        this.x = 0;
        this.bParam_ = 0;
        this.expr_ = '';
        Util.setErrorHandler();
        Util.setImagesDir(elem_ids.images_dir);
        if (typeof elem_ids.b_value === 'number') {
            this.bParam_ = elem_ids.b_value;
        }
        this.graphDiv = Util.getElementById(elem_ids.graph_div);
        const canvas = document.createElement('canvas');
        this.simCanvas = new LabCanvas(canvas, 'canvas1');
        this.simCanvas.setSize(800, 800);
        this.graphDiv.appendChild(canvas);
        this.graph_controls = Util.getElementById(elem_ids.graph_controls);
        const term_output = Util.maybeElementById(elem_ids.term_output);
        const term_input = Util.maybeElementById(elem_ids.term_input);
        this.terminal = new Terminal(term_input, term_output);
        Terminal.stdRegex(this.terminal);
        this.vars = new VarsList(['x', 'y'], ['x', 'y']);
        this.simRect = new DoubleRect(-10, -10, 10, 10);
        this.simView = new SimView('graphView', this.simRect);
        this.simView.setHorizAlign("FULL");
        this.simView.setVerticalAlign("FULL");
        this.simCanvas.addView(this.simView);
        const screenrect = this.simView.getScreenRect();
        this.graphLine = new GraphLine('GRAPH_LINE', this.vars);
        this.graphLine.setXVariable(0);
        this.graphLine.setYVariable(1);
        this.graphLine.setHotSpotColor('');
        this.graph = new DisplayGraph(this.graphLine);
        this.graph.setScreenRect(screenrect);
        this.simView.getDisplayList().add(this.graph);
        this.axes = new DisplayAxes(this.simView.getCoordMap().screenToSimRect(screenrect));
        this.simView.getDisplayList().add(this.axes);
        new GenericObserver(this.simView, evt => {
            if (evt.nameEquals(SimView.SCREEN_RECT_CHANGED)) {
                this.graph.setScreenRect(this.simView.getScreenRect());
            }
            else if (evt.nameEquals(SimView.SIM_RECT_CHANGED)) {
                const screenrect = this.simView.getScreenRect();
                const simrect = this.simView.getCoordMap().screenToSimRect(screenrect);
                this.axes.setSimRect(simrect);
            }
        }, 'adjust when screen rect changes');
        this.simCtrl = new SimController(this.simCanvas);
        this.terminal.setAfterEval(() => {
            this.graph.reset();
            this.simCanvas.paint();
        });
        let pn;
        this.addParameter(pn = new ParameterNumber(this, GraphCalc2App.en.B_PARAM, GraphCalc2App.i18n.B_PARAM, () => this.getBParam(), a => this.setBParam(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(3);
        const b_max = typeof elem_ids.b_max === 'number' ? elem_ids.b_max : 10;
        const b_min = typeof elem_ids.b_min === 'number' ? elem_ids.b_min : -10;
        const b_multiply = typeof elem_ids.b_multiply === 'boolean' ? elem_ids.b_multiply : false;
        const b_increments = typeof elem_ids.b_increments === 'number' ? elem_ids.b_increments : 200;
        this.addControl(new SliderControl(pn, b_min, b_max, b_multiply, b_increments));
        this.easyScript = new EasyScriptParser([this, this.simCanvas]);
        this.terminal.setParser(this.easyScript);
    }
    ;
    getClassName() {
        return 'GraphCalc2App';
    }
    ;
    addControl(control) {
        if (this.graph_controls != null) {
            const element = control.getElement();
            element.style.display = 'block';
            this.graph_controls.appendChild(element);
        }
        return control;
    }
    ;
    defineNames(myName) {
        this.terminal.addRegex('simCanvas|vars|terminal|graph|graphLine|axes|plot'
            + '|simCtrl|simView|x', myName + '.');
        this.terminal.addRegex2(new RegExp('\\bb\\b', 'g'), myName + '.getBParam()');
    }
    ;
    eval(script, output = true) {
        return this.terminal.eval(script, output);
    }
    ;
    plot(expr, range1, range2, numPts) {
        expr = this.expr_ = expr || this.expr_;
        numPts = numPts || 1000;
        const r = this.simView.getSimRect();
        range1 = range1 ?? r.getLeft();
        range2 = range2 ?? r.getRight();
        const incr = (range2 - range1) / numPts;
        if (incr <= 0) {
            this.terminal.println('ERROR calling plot: range2 must be greater than range1');
            return;
        }
        numPts += 1;
        this.vars.incrSequence(0);
        this.vars.incrSequence(1);
        this.x = range1;
        let y = NaN;
        if (expr) {
            for (let i = 0; i < numPts; i++) {
                y = this.terminal.eval(expr, false);
                if (typeof y === 'number' && isFinite(y)) {
                    this.vars.setValue(0, this.x, true);
                    this.vars.setValue(1, y, true);
                    this.graphLine.memorize();
                    this.x += incr;
                }
                else {
                    this.terminal.println('ERROR in plot: result is not a number ' + expr);
                    return;
                }
            }
        }
        this.graph.reset();
        this.simCanvas.paint();
    }
    ;
    setup() {
        this.terminal.parseURL();
    }
    ;
    start() {
    }
    ;
    getBParam() {
        return this.bParam_;
    }
    ;
    setBParam(value) {
        if (Util.veryDifferent(value, this.bParam_)) {
            this.bParam_ = value;
            this.graphLine.reset();
            this.plot();
            this.broadcastParameter(GraphCalc2App.en.B_PARAM);
        }
    }
    ;
}
GraphCalc2App.en = {
    B_PARAM: 'b'
};
GraphCalc2App.de_strings = {
    B_PARAM: 'b'
};
GraphCalc2App.i18n = Util.LOCALE === 'de' ? GraphCalc2App.de_strings : GraphCalc2App.en;
Util.defineGlobal('sims$experimental$GraphCalc2App', GraphCalc2App);
