import { AbstractApp } from '../common/AbstractApp.js';
import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { DisplayLine } from '../../lab/view/DisplayLine.js';
import { DisplayMagnetWheel } from './DisplayMagnetWheel.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { Force } from '../../lab/model/Force.js';
import { MagnetWheelSim } from './MagnetWheelSim.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { SimList } from '../../lab/model/SimList.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class MagnetWheelApp extends AbstractApp {
    constructor(elem_ids, opt_name) {
        Util.setErrorHandler();
        const simRect = new DoubleRect(-3, -2, 3, 2);
        const sim = new MagnetWheelSim();
        sim.getVarsList().setValue(1, 3);
        const advance = new SimpleAdvance(sim);
        advance.setTimeStep(0.01);
        super(elem_ids, simRect, sim, advance, sim, sim, opt_name);
        this.numMagnets_ = 12;
        this.symmetric_ = true;
        this.magnetAngle_ = 25;
        this.magnetWheel = sim.getMagnetWheel();
        this.magnetWheel.setMass(5);
        this.makeMagnets();
        this.dispWheel = new DisplayMagnetWheel(this.magnetWheel);
        this.displayList.add(this.dispWheel);
        this.sim.getSimList().addObserver(this);
        this.addPlaybackControls();
        let pn;
        pn = sim.getParameterNumber(MagnetWheelSim.en.MASS);
        this.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(MagnetWheelSim.en.DAMPING);
        this.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(MagnetWheelSim.en.MAGNET_STRENGTH);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, MagnetWheelSim.en.NUM_MAGNETS, MagnetWheelSim.i18n.NUM_MAGNETS, () => this.getNumMagnets(), a => this.setNumMagnets(a))
            .setDecimalPlaces(0));
        this.addControl(new SliderControl(pn, 1, 12, false));
        let pb;
        this.addParameter(pb = new ParameterBoolean(this, MagnetWheelSim.en.SYMMETRIC, MagnetWheelSim.i18n.SYMMETRIC, () => this.getSymmetric(), a => this.setSymmetric(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pn = new ParameterNumber(this, MagnetWheelSim.en.MAGNET_ANGLE, MagnetWheelSim.i18n.MAGNET_ANGLE, () => this.getMagnetAngle(), a => this.setMagnetAngle(a))
            .setDecimalPlaces(1));
        this.magnetAngleControl_ = new NumericControl(pn);
        this.magnetAngleControl_.setEnabled(!this.symmetric_);
        this.addControl(this.magnetAngleControl_);
        pn = sim.getParameterNumber(EnergyInfo.en.PE_OFFSET);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dispWheel: ' + this.dispWheel.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'MagnetWheelApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('magnetWheel|dispWheel', myName + '.');
    }
    ;
    observe(event) {
        if (event.getSubject() == this.sim.getSimList()) {
            const obj = event.getValue();
            if (event.nameEquals(SimList.OBJECT_ADDED)) {
                if (this.displayList.find(obj)) {
                    return;
                }
                if (obj instanceof Force) {
                    const dl = new DisplayLine(obj).setThickness(1);
                    dl.setColor('blue');
                    dl.setZIndex(10);
                    this.displayList.add(dl);
                }
            }
            else if (event.nameEquals(SimList.OBJECT_REMOVED)) {
                const d = this.displayList.find(obj);
                if (d) {
                    this.displayList.remove(d);
                }
            }
        }
    }
    ;
    getNumMagnets() {
        return this.numMagnets_;
    }
    ;
    setNumMagnets(value) {
        value = Math.floor(value);
        if (value < 1 || value > 12) {
            throw 'number of magnets must be from 1 to 12';
        }
        this.numMagnets_ = value;
        this.makeMagnets();
        this.broadcastParameter(MagnetWheelSim.en.NUM_MAGNETS);
    }
    ;
    getSymmetric() {
        return this.symmetric_;
    }
    ;
    setSymmetric(value) {
        this.symmetric_ = value;
        this.makeMagnets();
        this.magnetAngleControl_.setEnabled(!this.symmetric_);
        this.broadcastParameter(MagnetWheelSim.en.SYMMETRIC);
    }
    ;
    getMagnetAngle() {
        return this.magnetAngle_;
    }
    ;
    setMagnetAngle(value) {
        if (value < 1 || value > 360) {
            throw 'magnet angle must be from 1 to 360';
        }
        this.magnetAngle_ = value;
        this.makeMagnets();
        this.broadcastParameter(MagnetWheelSim.en.MAGNET_ANGLE);
    }
    ;
    makeMagnets() {
        const magnets = [];
        const n = this.numMagnets_;
        const r = this.magnetWheel.getRadius() * 0.85;
        for (let i = 0; i < n; i++) {
            const a = i * 2 * Math.PI * (this.symmetric_ ? 1 / n : this.magnetAngle_ / 360);
            magnets.push(new Vector(r * Math.cos(a), r * Math.sin(a)));
        }
        this.magnetWheel.setMagnets(magnets);
    }
    ;
}
Util.defineGlobal('sims$misc$MagnetWheelApp', MagnetWheelApp);
