import { AbstractApp } from '../common/AbstractApp.js';
import { DisplayLine } from '../../lab/view/DisplayLine.js';
import { DisplayRobotWheel } from './DisplayRobotWheel.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Force } from '../../lab/model/Force.js';
import { GenericMemo } from '../../lab/util/Memo.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { RobotSpeedSim } from './RobotSpeedSim.js';
import { SimList } from '../../lab/model/SimList.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { TabLayout3 } from '../common/TabLayout3.js';
import { Util } from '../../lab/util/Util.js';
export class RobotSpeedApp extends AbstractApp {
    constructor(elem_ids, opt_name) {
        Util.setErrorHandler();
        const simRect = new DoubleRect(-0.7, -0.4, 2.5, 1.3);
        const sim = new RobotSpeedSim();
        const advance = new SimpleAdvance(sim);
        super(elem_ids, simRect, sim, advance, null, null, opt_name);
        this.ramp = new DisplayShape(this.simList.getPointMass('ramp'));
        this.ramp.setFillStyle('black');
        this.displayList.add(this.ramp);
        const bot = this.simList.getPointMass('robot');
        this.robot = new DisplayShape(bot);
        this.robot.setFillStyle('lightGray');
        this.robot.setDrawCenterOfMass(true);
        this.displayList.add(this.robot);
        this.wheelf = new DisplayRobotWheel(this.simList.getPointMass('wheelf'));
        this.displayList.add(this.wheelf);
        this.wheelr = new DisplayRobotWheel(this.simList.getPointMass('wheelr'));
        this.displayList.add(this.wheelr);
        this.addPlaybackControls();
        let pn = sim.getParameterNumber(RobotSpeedSim.en.DIAMETER);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.MASS);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.TORQUE);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.FREE_SPEED);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.SLOPE);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.COEF_FRICTION);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(RobotSpeedSim.en.CENTER_OF_MASS);
        this.addControl(new NumericControl(pn));
        this.graph.addControl(new NumericControl(pn));
        this.timeGraph.addControl(new NumericControl(pn));
        this.memo = new GenericMemo(() => {
            const p = bot.getPosition().getX();
            if (p < -0.5 || p > 6)
                this.simRun.pause();
        });
        this.simRun.addMemo(this.memo);
        this.simRun.setTimeStep(0.01);
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.graph.line.setXVariable(2);
        this.graph.line.setYVariable(1);
        this.timeGraph.line1.setYVariable(1);
        this.timeGraph.line2.setYVariable(4);
        this.timeGraph.line3.setYVariable(5);
        this.timeGraph.autoScale.setTimeWindow(2);
        this.sim.getSimList().addObserver(this);
        this.getParameterBoolean('PAN_ZOOM').setValue(true);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', robot: ' + this.robot.toStringShort()
            + ', ramp: ' + this.ramp.toStringShort()
            + ', wheelf: ' + this.wheelf.toStringShort()
            + ', wheelr: ' + this.wheelr.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'RobotSpeedApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('body|ramp|wheelf|wheelr', myName + '.');
    }
    ;
    makeLayout(elem_ids) {
        const canvasWidth = 800;
        const canvasHeight = Math.round(canvasWidth * this.simRect.getHeight() /
            this.simRect.getWidth());
        const layout = new TabLayout3(elem_ids, canvasWidth, canvasHeight);
        layout.setLayout("sim+time_graph");
        return layout;
    }
    ;
    observe(event) {
        if (event.getSubject() == this.sim.getSimList()) {
            const obj = event.getValue();
            if (event.nameEquals(SimList.OBJECT_ADDED)) {
                if (this.displayList.find(obj)) {
                    return;
                }
                if (obj instanceof Force) {
                    const dl = new DisplayLine(obj).setThickness(1);
                    dl.setColor('blue');
                    dl.setZIndex(10);
                    this.displayList.add(dl);
                }
            }
            else if (event.nameEquals(SimList.OBJECT_REMOVED)) {
                const d = this.displayList.find(obj);
                if (d) {
                    this.displayList.remove(d);
                }
            }
        }
    }
    ;
}
Util.defineGlobal('sims$misc$RobotSpeedApp', RobotSpeedApp);
