import { AbstractSubject } from '../../lab/util/AbstractSubject.js';
import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { ChoiceControl } from '../../lab/controls/ChoiceControl.js';
import { Clock } from '../../lab/util/Clock.js';
import { CommonControls } from '../common/CommonControls.js';
import { DisplayClock } from '../../lab/view/DisplayClock.js';
import { DisplayPath } from '../../lab/view/DisplayPath.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DisplayText } from '../../lab/view/DisplayText.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { DrawingStyle } from '../../lab/view/DrawingStyle.js';
import { EnergyBarGraph } from '../../lab/graph/EnergyBarGraph.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { FlatShape, TrianglePulseShape, SquarePulseShape, SinePulseShape, HalfSinePulseShape, MultiSineShape } from './StringShapes.js';
import { GenericEvent } from '../../lab/util/Observe.js';
import { GenericMemo } from '../../lab/util/Memo.js';
import { GenericObserver } from '../../lab/util/Observe.js';
import { LabelControl } from '../../lab/controls/LabelControl.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterString } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { SimController } from '../../lab/app/SimController.js';
import { SimRunner } from '../../lab/app/SimRunner.js';
import { SimView } from '../../lab/view/SimView.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { StringAdvance } from './StringAdvance.js';
import { StringSim, StringPath } from './StringSim.js';
import { TabLayout } from '../common/TabLayout.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class StringApp extends AbstractSubject {
    constructor(elem_ids) {
        Util.setErrorHandler();
        super('APP');
        this.stability = -1;
        this.layout = new TabLayout(elem_ids);
        this.terminal = this.layout.getTerminal();
        const sim_controls = this.layout.getSimControls();
        const simCanvas = this.layout.getSimCanvas();
        const length = 13.5;
        this.shapes = [
            new FlatShape(length),
            new TrianglePulseShape(length),
            new SquarePulseShape(length),
            new SinePulseShape(length),
            new HalfSinePulseShape(length),
            new MultiSineShape(length)
        ];
        this.sim = new StringSim(this.shapes[1]);
        this.terminal.setAfterEval(() => this.sim.modifyObjects());
        this.simList = this.sim.getSimList();
        this.simCtrl = new SimController(simCanvas, this.sim);
        this.advance = new StringAdvance(this.sim);
        this.simRect = new DoubleRect(-1, -0.25, 14, 0.25);
        this.simView = new SimView('simView', this.simRect);
        this.displayList = this.simView.getDisplayList();
        this.simView.setHorizAlign("FULL");
        this.simView.setVerticalAlign("FULL");
        simCanvas.addView(this.simView);
        this.statusView = new SimView('status', new DoubleRect(-10, -10, 10, 10));
        simCanvas.addView(this.statusView);
        this.axes = CommonControls.makeAxes(this.simView, true);
        this.simRun = new SimRunner(this.advance);
        this.simRun.addCanvas(simCanvas);
        this.clock = this.simRun.getClock();
        this.blockMass = this.simList.getPointMass('block');
        this.block = new DisplayShape(this.blockMass);
        this.block.setFillStyle('rgba(0,0,255,0.2)');
        this.displayList.add(this.block);
        this.blockText = new DisplayText('drag');
        this.blockText.setFillStyle('rgba(255,255,255,0.7)');
        this.blockText.setTextAlign('center');
        this.statusView.getDisplayList().add(this.blockText);
        this.statusView.addMemo(new GenericMemo(() => {
            const map1 = this.simView.getCoordMap();
            const map2 = this.statusView.getCoordMap();
            const loc = map1.simToScreen(this.blockMass.getPosition());
            this.blockText.setPosition(map2.screenToSim(loc));
        }, 'blockText follows blockMass'));
        this.shadow = PointMass.makeSquare(1, 'shadow');
        this.showShadow = new DisplayShape(this.shadow);
        this.showShadow.setFillStyle('');
        this.showShadow.setStrokeStyle('gray');
        this.statusView.getDisplayList().add(this.showShadow);
        this.statusView.addMemo(new GenericMemo(() => {
            const map1 = this.simView.getCoordMap();
            const map2 = this.statusView.getCoordMap();
            const w = this.blockMass.getWidth();
            const w2 = w * map1.getScaleX() / map2.getScaleX();
            this.shadow.setWidth(w2);
            const h = this.blockMass.getHeight();
            const h2 = h * map1.getScaleY() / map2.getScaleY();
            this.shadow.setHeight(h2);
            const loc = map1.simToScreen(this.blockMass.getPosition());
            this.shadow.setPosition(map2.screenToSim(loc));
        }, 'shadow outline follows blockMass'));
        this.stabilityText = new DisplayText();
        this.stabilityText.setPosition(new Vector(-5, -7));
        this.statusView.getDisplayList().add(this.stabilityText);
        new GenericObserver(this.sim, _evt => {
            this.sim.modifyObjects();
            const s = this.sim.getStability();
            if (this.stability != s) {
                this.stabilityText.setText('stability = ' + Util.NF5(s));
                this.stability = s;
                this.stabilityText.setFillStyle(s < 1 ? 'rgb(160,160,160)' : 'red');
            }
        }, 'modifyObjects after parameter or variable change');
        this.sim.broadcast(new GenericEvent(this.sim, 'event'));
        this.path = new StringPath(this.sim);
        this.displayPath = new DisplayPath();
        this.displayPath.setScreenRect(this.simView.getScreenRect());
        this.displayPath.setUseBuffer(false);
        this.displayPath.setZIndex(-1);
        this.displayList.add(this.displayPath);
        this.displayPath.addPath(this.path, DrawingStyle.dotStyle('red', 2));
        this.addControl(CommonControls.makePlaybackControls(this.simRun));
        const sn = [];
        const snl = [];
        for (let i = 0; i < this.shapes.length; i++) {
            sn.push(this.shapes[i].getName(false));
            snl.push(this.shapes[i].getName(true));
        }
        const ps = new ParameterString(this, StringSim.en.SHAPE, StringSim.i18n.SHAPE, () => this.getShape(), a => this.setShape(a), snl, sn);
        this.addParameter(ps);
        this.addControl(new ChoiceControl(ps));
        let pn = this.sim.getParameterNumber(StringSim.en.NUM_POINTS);
        this.addControl(new NumericControl(pn));
        pn = this.sim.getParameterNumber(StringSim.en.DENSITY);
        this.addControl(new SliderControl(pn, 0.2, 20.2, true));
        pn = this.sim.getParameterNumber(StringSim.en.TENSION);
        this.addControl(new SliderControl(pn, 0, 20, false));
        pn = this.sim.getParameterNumber(StringSim.en.DAMPING);
        this.addControl(new SliderControl(pn, 0, 1, false));
        pn = this.sim.getParameterNumber(StringSim.en.TIME_STEP);
        let nc = new NumericControl(pn);
        nc.setDecimalPlaces(7);
        this.addControl(nc);
        pn = this.clock.getParameterNumber(Clock.en.TIME_RATE);
        this.addControl(new NumericControl(pn));
        pn = this.sim.getParameterNumber(EnergyInfo.en.PE_OFFSET);
        this.addControl(new NumericControl(pn));
        if (0 == 1 && Util.DEBUG) {
            this.addControl(new LabelControl('compiled ' + Util.COMPILE_TIME));
        }
        this.energyGraph = new EnergyBarGraph(this.sim);
        this.showEnergyParam = CommonControls.makeShowEnergyParam(this.energyGraph, this.statusView, this);
        this.addControl(new CheckBoxControl(this.showEnergyParam));
        this.displayClock = new DisplayClock(() => this.sim.getTime(), () => this.clock.getRealTime(), 2, 2);
        this.displayClock.setPosition(new Vector(8, 4));
        let pb = CommonControls.makeShowClockParam(this.displayClock, this.statusView, this);
        this.addControl(new CheckBoxControl(pb));
        const panzoom = CommonControls.makePanZoomControls(this.simView, true, () => this.simView.setSimRect(this.simRect));
        this.layout.getSimDiv().appendChild(panzoom);
        pb = CommonControls.makeShowPanZoomParam(panzoom, this);
        pb.setValue(false);
        this.addControl(new CheckBoxControl(pb));
        const bm = CommonControls.makeBackgroundMenu(this.layout.getSimCanvas());
        this.addControl(bm);
        new GenericObserver(this.sim, evt => {
            if (evt.nameEquals(StringSim.en.TIME_STEP)) {
                this.simRun.setTimeStep(this.sim.getTimeStep());
            }
        }, 'keep SimRunner\'s timeStep same as Simulation\'s');
        let subjects = [
            this,
            this.sim,
            this.simRun,
            this.clock,
            this.simView,
            this.statusView
        ];
        subjects = subjects.concat(this.layout.getSubjects());
        this.easyScript = CommonControls.makeEasyScript(subjects, [], this.simRun, this.terminal);
        this.addControl(CommonControls.makeURLScriptButton(this.easyScript, this.simRun));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', sim: ' + this.sim.toStringShort()
            + ', simRun: ' + this.simRun.toStringShort()
            + ', simView: ' + this.simView.toStringShort()
            + ', statusView: ' + this.statusView.toStringShort()
            + ', terminal: ' + this.terminal
            + ', path: ' + this.path
            + ', displayPath: ' + this.displayPath
            + ', shapes: [ ' + this.shapes + ']'
            + super.toString();
    }
    ;
    getClassName() {
        return 'StringApp';
    }
    ;
    getShape() {
        return this.sim.getShape().getName();
    }
    setShape(value) {
        for (let i = 0; i < this.shapes.length; i++) {
            const shape = this.shapes[i];
            if (shape.getName() == value) {
                this.sim.setShape(shape);
                this.broadcastParameter(StringSim.en.SHAPE);
                return;
            }
        }
        throw 'unknown shape ' + value;
    }
    defineNames(myName) {
        this.terminal.addRegex('advance|axes|block|blockMass|clock|displayClock|energyGraph'
            + '|path|displayPath|displayList'
            + '|layout|sim|simCtrl|simList|simRun|simView|statusView|terminal|easyScript', myName + '.');
        this.terminal.addRegex('simCanvas', myName + '.layout.');
    }
    ;
    addControl(control) {
        return this.layout.addControl(control);
    }
    ;
    setup() {
        this.clock.resume();
        this.terminal.parseURL();
        this.sim.saveInitialState();
        this.sim.modifyObjects();
    }
    ;
    start() {
        this.simRun.startFiring();
    }
    ;
    eval(script, output = true) {
        return this.terminal.eval(script, output);
    }
    ;
}
Util.defineGlobal('sims$pde$StringApp', StringApp);
