import { AbstractODESim } from '../../lab/model/ODESim.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TH1 = 0;
const TH1P = 1;
const TH2 = 2;
const TH2P = 3;
const TH1PP = 4;
const TH2PP = 5;
const KE = 6;
const PE = 7;
const TE = 8;
const TIME = 9;
export class DoublePendulumSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.rod1Length_ = 1;
        this.rod2Length_ = 1;
        this.rod1_ = new ConcreteLine('rod1');
        this.rod2_ = new ConcreteLine('rod2');
        this.bob1_ = PointMass.makeCircle(0.2, 'bob1');
        this.bob2_ = PointMass.makeCircle(0.2, 'bob2');
        this.gravity_ = 9.8;
        this.potentialOffset_ = 0;
        this.isDragging_ = false;
        const var_names = [
            DoublePendulumSim.en.ANGLE_1,
            DoublePendulumSim.en.ANGLE_1_VELOCITY,
            DoublePendulumSim.en.ANGLE_2,
            DoublePendulumSim.en.ANGLE_2_VELOCITY,
            DoublePendulumSim.en.ACCELERATION_1,
            DoublePendulumSim.en.ACCELERATION_2,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME
        ];
        const i18n_names = [
            DoublePendulumSim.i18n.ANGLE_1,
            DoublePendulumSim.i18n.ANGLE_1_VELOCITY,
            DoublePendulumSim.i18n.ANGLE_2,
            DoublePendulumSim.i18n.ANGLE_2_VELOCITY,
            DoublePendulumSim.i18n.ACCELERATION_1,
            DoublePendulumSim.i18n.ACCELERATION_2,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(TH1PP, TH2PP, KE, PE, TE);
        this.bob1_.setMass(2.0);
        this.bob2_.setMass(2.0);
        this.getSimList().add(this.rod1_, this.rod2_, this.bob1_, this.bob2_);
        this.restState();
        this.getVarsList().setValue(0, Math.PI / 8);
        this.saveInitialState();
        this.addParameter(new ParameterNumber(this, DoublePendulumSim.en.ROD_1_LENGTH, DoublePendulumSim.i18n.ROD_1_LENGTH, () => this.getRod1Length(), a => this.setRod1Length(a)));
        this.addParameter(new ParameterNumber(this, DoublePendulumSim.en.ROD_2_LENGTH, DoublePendulumSim.i18n.ROD_2_LENGTH, () => this.getRod2Length(), a => this.setRod2Length(a)));
        this.addParameter(new ParameterNumber(this, DoublePendulumSim.en.MASS_1, DoublePendulumSim.i18n.MASS_1, () => this.getMass1(), a => this.setMass1(a)));
        this.addParameter(new ParameterNumber(this, DoublePendulumSim.en.MASS_2, DoublePendulumSim.i18n.MASS_2, () => this.getMass2(), a => this.setMass2(a)));
        this.addParameter(new ParameterNumber(this, DoublePendulumSim.en.GRAVITY, DoublePendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        let pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
        this.addParameter(pn);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', rod1Length_: ' + Util.NF(this.rod1Length_)
            + ', rod2Length_: ' + Util.NF(this.rod2Length_)
            + ', rod1_: ' + this.rod1_
            + ', rod2_: ' + this.rod2_
            + ', bob1_: ' + this.bob1_
            + ', bob2_: ' + this.bob2_
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'DoublePendulumSim';
    }
    ;
    restState() {
        const va = this.getVarsList();
        const vars = va.getValues();
        vars[TH1] = 0;
        vars[TH1P] = 0;
        vars[TH2] = 0;
        vars[TH2P] = 0;
        va.setValues(vars);
        this.moveObjects(vars);
    }
    ;
    getEnergyInfo() {
        const L1 = this.rod1Length_;
        const L2 = this.rod2Length_;
        const ke = this.bob1_.getKineticEnergy() + this.bob2_.getKineticEnergy();
        const y1 = this.bob1_.getPosition().getY();
        const y2 = this.bob2_.getPosition().getY();
        const pe = this.gravity_ * this.bob1_.getMass() * (y1 - -L1)
            + this.gravity_ * this.bob2_.getMass() * (y2 - (-L1 - L2));
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const theta1 = Util.limitAngle(vars[TH1]);
        if (theta1 != vars[TH1]) {
            this.getVarsList().setValue(0, theta1, false);
            vars[TH1] = theta1;
        }
        const theta2 = Util.limitAngle(vars[TH2]);
        if (theta2 != vars[TH2]) {
            this.getVarsList().setValue(2, theta2, false);
            vars[TH2] = theta2;
        }
        this.moveObjects(vars);
        const rate = new Array(vars.length);
        this.evaluate(vars, rate, 0);
        vars[TH1PP] = rate[TH1P];
        vars[TH2PP] = rate[TH2P];
        const ei = this.getEnergyInfo();
        vars[KE] = ei.getTranslational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        va.setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        const theta1 = vars[TH1];
        const sinTheta1 = Math.sin(theta1);
        const cosTheta1 = Math.cos(theta1);
        const theta2 = vars[TH2];
        const sinTheta2 = Math.sin(theta2);
        const cosTheta2 = Math.cos(theta2);
        const L1 = this.rod1Length_;
        const L2 = this.rod2Length_;
        const x1 = L1 * sinTheta1;
        const y1 = -L1 * cosTheta1;
        const x2 = x1 + L2 * sinTheta2;
        const y2 = y1 - L2 * cosTheta2;
        this.bob1_.setPosition(new Vector(x1, y1));
        this.bob2_.setPosition(new Vector(x2, y2));
        const v1x = vars[TH1P] * L1 * cosTheta1;
        const v1y = vars[TH1P] * L1 * sinTheta1;
        const v2x = v1x + vars[TH2P] * L2 * cosTheta2;
        const v2y = v1y + vars[TH2P] * L2 * sinTheta2;
        this.bob1_.setVelocity(new Vector(v1x, v1y));
        this.bob2_.setVelocity(new Vector(v2x, v2y));
        this.rod1_.setStartPoint(Vector.ORIGIN);
        this.rod1_.setEndPoint(this.bob1_.getPosition());
        this.rod2_.setStartPoint(this.bob1_.getPosition());
        this.rod2_.setEndPoint(this.bob2_.getPosition());
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.bob1_ || simObject == this.bob2_) {
            this.isDragging_ = true;
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = location.subtract(offset);
        if (simObject == this.bob1_) {
            vars[TH1] = Math.atan2(p.getX(), -p.getY());
            vars[TH1P] = 0;
            vars[TH2P] = 0;
        }
        else if (simObject == this.bob2_) {
            vars[TH2] = Math.atan2(p.getX() - this.bob1_.getPosition().getX(), -(p.getY() - this.bob1_.getPosition().getY()));
            vars[TH1P] = 0;
            vars[TH2P] = 0;
        }
        va.setValues(vars);
        this.moveObjects(vars);
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging_ = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (!this.isDragging_) {
            const th1 = vars[TH1];
            const dth1 = vars[TH1P];
            const th2 = vars[TH2];
            const dth2 = vars[TH2P];
            const m2 = this.bob2_.getMass();
            const m1 = this.bob1_.getMass();
            const L1 = this.rod1Length_;
            const L2 = this.rod2Length_;
            const g = this.gravity_;
            change[TH1] = dth1;
            let num = -g * (2 * m1 + m2) * Math.sin(th1);
            num = num - g * m2 * Math.sin(th1 - 2 * th2);
            num = num - 2 * m2 * dth2 * dth2 * L2 * Math.sin(th1 - th2);
            num = num - m2 * dth1 * dth1 * L1 * Math.sin(2 * (th1 - th2));
            num = num / (L1 * (2 * m1 + m2 - m2 * Math.cos(2 * (th1 - th2))));
            change[TH1P] = num;
            change[TH2] = dth2;
            num = (m1 + m2) * dth1 * dth1 * L1;
            num = num + g * (m1 + m2) * Math.cos(th1);
            num = num + m2 * dth2 * dth2 * L2 * Math.cos(th1 - th2);
            num = num * 2 * Math.sin(th1 - th2);
            num = num / (L2 * (2 * m1 + m2 - m2 * Math.cos(2 * (th1 - th2))));
            change[TH2P] = num;
        }
        return null;
    }
    ;
    getMass1() {
        return this.bob1_.getMass();
    }
    ;
    setMass1(value) {
        this.bob1_.setMass(value);
        this.getVarsList().incrSequence(TH1PP, TH2PP, KE, PE, TE);
        this.broadcastParameter(DoublePendulumSim.en.MASS_1);
    }
    ;
    getMass2() {
        return this.bob2_.getMass();
    }
    ;
    setMass2(value) {
        this.bob2_.setMass(value);
        this.getVarsList().incrSequence(TH1PP, TH2PP, KE, PE, TE);
        this.broadcastParameter(DoublePendulumSim.en.MASS_2);
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(TH1PP, TH2PP, PE, TE);
        this.broadcastParameter(DoublePendulumSim.en.GRAVITY);
    }
    ;
    getRod1Length() {
        return this.rod1Length_;
    }
    ;
    setRod1Length(value) {
        this.rod1Length_ = value;
        this.getVarsList().incrSequence(TH1PP, TH2PP, KE, PE, TE);
        this.broadcastParameter(DoublePendulumSim.en.ROD_1_LENGTH);
    }
    ;
    getRod2Length() {
        return this.rod2Length_;
    }
    ;
    setRod2Length(value) {
        this.rod2Length_ = value;
        this.getVarsList().incrSequence(TH1PP, TH2PP, KE, PE, TE);
        this.broadcastParameter(DoublePendulumSim.en.ROD_2_LENGTH);
    }
    ;
}
DoublePendulumSim.en = {
    ACCELERATION_1: 'acceleration-1',
    ACCELERATION_2: 'acceleration-2',
    ANGLE_1: 'angle-1',
    ANGLE_1_VELOCITY: 'angle-1 velocity',
    ANGLE_2: 'angle-2',
    ANGLE_2_VELOCITY: 'angle-2 velocity',
    GRAVITY: 'gravity',
    MASS_1: 'mass-1',
    MASS_2: 'mass-2',
    ROD_1_LENGTH: 'rod-1 length',
    ROD_2_LENGTH: 'rod-2 length'
};
DoublePendulumSim.de_strings = {
    ACCELERATION_1: 'Beschleunigung-1',
    ACCELERATION_2: 'Beschleunigung-2',
    ANGLE_1: 'Winkel-1',
    ANGLE_1_VELOCITY: 'Winkel-1 Geschwindigkeit',
    ANGLE_2: 'Winkel-2',
    ANGLE_2_VELOCITY: 'Winkel-2 Geschwindigkeit',
    GRAVITY: 'Gravitation',
    MASS_1: 'Masse-1',
    MASS_2: 'Masse-2',
    ROD_1_LENGTH: 'Stange-1 Länge',
    ROD_2_LENGTH: 'Stange-2 Länge'
};
DoublePendulumSim.i18n = Util.LOCALE === 'de' ? DoublePendulumSim.de_strings : DoublePendulumSim.en;
Util.defineGlobal('sims$pendulum$DoublePendulumSim', DoublePendulumSim);
