import { AbstractODESim } from '../../lab/model/ODESim.js';
import { GenericEvent } from '../../lab/util/Observe.js';
import { NumericalPath } from '../../lab/model/NumericalPath.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PathPoint } from '../../lab/model/PathPoint.js';
import { PointMass, ShapeType } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TIME = 0;
const P0 = 1;
const V0 = 2;
const X0 = 3;
const Y0 = 4;
export class BrachistoSim extends AbstractODESim {
    constructor(paths, opt_name, opt_simList) {
        super(opt_name, opt_simList);
        this.simObjects_ = [];
        this.damping_ = 0;
        this.gravity_ = 2.0;
        this.mass_ = 1.0;
        this.choice_ = -1;
        const plen = paths.length;
        this.balls_ = new Array(plen);
        this.paths_ = new Array(plen);
        for (let i = 0; i < plen; i++) {
            this.paths_[i] = new NumericalPath(paths[i]);
            this.getSimList().add(this.paths_[i]);
        }
        this.setVarsList(new VarsList(BrachistoSim.makeVarNames(paths, false), BrachistoSim.makeVarNames(paths, true), this.getName() + '_VARS'));
        const cv = this.getVarsList().toArray().map((v) => {
            const j = 1 + (v.indexOf() - 1) % 4;
            if (j == X0 || j == Y0) {
                v.setComputed(true);
            }
        });
        this.addParameter(new ParameterNumber(this, BrachistoSim.en.DAMPING, BrachistoSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, BrachistoSim.en.GRAVITY, BrachistoSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, BrachistoSim.en.MASS, BrachistoSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', mass_: ' + Util.NF(this.mass_)
            + ', paths: ' + this.paths_.length
            + super.toString();
    }
    ;
    getClassName() {
        return 'BrachistoSim';
    }
    ;
    static makeVarNames(paths, localized) {
        const names = new Array(paths.length * 4 + 1);
        for (let i = 0, len = names.length; i < len; i++) {
            names[i] = BrachistoSim.getVariableName(i, paths, localized);
        }
        return names;
    }
    ;
    static getVariableName(i, paths, localized) {
        if (i == 0) {
            return localized ? VarsList.i18n.TIME : VarsList.en.TIME;
        }
        const j = 1 + (i - 1) % 4;
        const obj = Math.floor((i - 1) / 4);
        const name = paths[obj].getName();
        let type;
        if (j == V0) {
            type = localized ? BrachistoSim.i18n.VELOCITY : BrachistoSim.en.VELOCITY;
        }
        else {
            type = localized ? BrachistoSim.i18n.POSITION : BrachistoSim.en.POSITION;
        }
        switch (j) {
            case P0:
            case V0: return name + ' ' + type;
            case X0: return name + ' x-' + type;
            case Y0: return name + ' y-' + type;
            default: Util.assert(false);
        }
        return '';
    }
    ;
    getPaths() {
        return Array.from(this.paths_);
    }
    ;
    getPathChoice() {
        return this.choice_;
    }
    ;
    setPathChoice(choice) {
        this.choice_ = choice;
        this.getSimList().removeAll(this.simObjects_);
        this.simObjects_.length = 0;
        const va = this.getVarsList();
        va.setValue(TIME, 0);
        for (let i = 0, len = this.paths_.length; i < len; i++) {
            va.setValue(P0 + 4 * i, this.paths_[i].map_x_to_p(0.0));
            va.setValue(V0 + 4 * i, 0);
        }
        this.saveInitialState();
        let p = PointMass.makeCircle(0.2, 'start');
        p.setPosition(new Vector(0, 0));
        p.setShape(ShapeType.RECTANGLE);
        this.simObjects_.push(p);
        p = PointMass.makeCircle(0.2, 'end');
        p.setPosition(new Vector(3, -2));
        p.setShape(ShapeType.RECTANGLE);
        this.simObjects_.push(p);
        if (choice >= 0) {
            const len = this.balls_.length;
            for (let i = 0; i < len; i++) {
                const name = (i == choice) ? 'ball selected' : 'ball';
                p = this.balls_[i] = PointMass.makeCircle(0.1, name);
                this.simObjects_.push(p);
            }
        }
        this.getSimList().addAll(this.simObjects_);
        this.broadcast(new GenericEvent(this, BrachistoSim.PATH_CHOSEN, choice));
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        for (let i = 0, len = this.paths_.length; i < len; i++) {
            const pathPoint = new PathPoint(vars[P0 + 4 * i]);
            this.paths_[i].map_p_to_slope(pathPoint);
            const veloVector = pathPoint.getSlope().multiply(vars[V0 + 4 * i]);
            const b = this.balls_[i];
            if (b !== undefined) {
                b.setPosition(pathPoint);
                b.setVelocity(veloVector);
            }
            va.setValue(X0 + 4 * i, pathPoint.getX(), true);
            va.setValue(Y0 + 4 * i, pathPoint.getY(), true);
        }
    }
    ;
    startDrag(_simObject, location, _offset, _dragBody, _modifiers) {
        let dist = Infinity;
        let closestPath = -1;
        for (let i = 0, len = this.paths_.length; i < len; i++) {
            const pathPoint = this.paths_[i].findNearestGlobal(location);
            const d = pathPoint.getPosition().distanceTo(location);
            if (d < dist) {
                closestPath = i;
                dist = d;
            }
        }
        this.setPathChoice(closestPath);
        return true;
    }
    ;
    mouseDrag(_simObject, _location, _offset) {
    }
    ;
    finishDrag(_simObject, _location, _offset) {
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        change[TIME] = 1;
        for (let i = 0, len = this.balls_.length; i < len; i++) {
            const idx = 4 * i;
            let mass = 0;
            const b = this.balls_[i];
            if (b === undefined) {
                change[P0 + idx] = 0;
                change[V0 + idx] = 0;
                change[X0 + idx] = 0;
                change[Y0 + idx] = 0;
                continue;
            }
            else {
                mass = b.getMass();
            }
            change[P0 + idx] = vars[V0 + idx];
            const pathPoint = new PathPoint(vars[P0 + idx]);
            this.paths_[i].map_p_to_slope(pathPoint);
            const k = pathPoint.slope;
            const sinTheta = isFinite(k) ? k / Math.sqrt(1 + k * k) : 1;
            change[V0 + idx] = -this.gravity_ * pathPoint.direction * sinTheta;
            change[V0 + idx] -= this.damping_ * vars[V0 + idx] / mass;
            change[X0 + idx] = change[Y0 + idx] = 0;
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.broadcastParameter(BrachistoSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(BrachistoSim.en.DAMPING);
    }
    ;
    getMass() {
        return this.mass_;
    }
    ;
    setMass(value) {
        this.mass_ = value;
        this.balls_.forEach(b => { if (b !== undefined)
            b.setMass(value); });
        this.broadcastParameter(BrachistoSim.en.MASS);
    }
    ;
}
BrachistoSim.PATH_CHOSEN = 'PATH_CHOSEN';
BrachistoSim.en = {
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    MASS: 'mass',
    POSITION: 'position',
    VELOCITY: 'velocity'
};
BrachistoSim.de_strings = {
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    POSITION: 'Position',
    VELOCITY: 'Geschwindigkeit'
};
BrachistoSim.i18n = Util.LOCALE === 'de' ? BrachistoSim.de_strings : BrachistoSim.en;
Util.defineGlobal('sims$roller$BrachistoSim', BrachistoSim);
