import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PathPoint } from '../../lab/model/PathPoint.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TRACK_P = 0;
const TRACK_V = 1;
const TIME = 2;
const X = 3;
const Y = 4;
const KE = 5;
const PE = 6;
const TE = 7;
const ANCHOR_X = 8;
const ANCHOR_Y = 9;
export class RollerSingleSim extends AbstractODESim {
    constructor(hasSpring, opt_name) {
        super(opt_name);
        this.damping_ = 0;
        this.gravity_ = 9.8;
        this.path_ = null;
        this.lowestPoint_ = 0;
        this.potentialOffset_ = 0;
        this.pathPoint_ = new PathPoint();
        this.dragObj_ = null;
        const var_names = [
            RollerSingleSim.en.POSITION,
            RollerSingleSim.en.VELOCITY,
            VarsList.en.TIME,
            RollerSingleSim.en.X_POSITION,
            RollerSingleSim.en.Y_POSITION,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            RollerSingleSim.en.ANCHOR_X,
            RollerSingleSim.en.ANCHOR_Y
        ];
        const i18n_names = [
            RollerSingleSim.i18n.POSITION,
            RollerSingleSim.i18n.VELOCITY,
            VarsList.i18n.TIME,
            RollerSingleSim.i18n.X_POSITION,
            RollerSingleSim.i18n.Y_POSITION,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            RollerSingleSim.i18n.ANCHOR_X,
            RollerSingleSim.i18n.ANCHOR_Y
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(X, Y, KE, PE, TE);
        this.hasSpring_ = hasSpring || false;
        this.ball1_ = PointMass.makeCircle(0.3, 'ball1');
        this.ball1_.setMass(0.5);
        this.getSimList().add(this.ball1_);
        this.anchor_ = PointMass.makeSquare(0.4, 'anchor');
        this.spring_ = new Spring('spring', this.ball1_, Vector.ORIGIN, this.anchor_, Vector.ORIGIN, 1.0, 5.0);
        if (this.hasSpring_) {
            this.getSimList().add(this.anchor_, this.spring_);
        }
        let pn;
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.DAMPING, RollerSingleSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.GRAVITY, RollerSingleSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.MASS, RollerSingleSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.SPRING_DAMPING, RollerSingleSim.i18n.SPRING_DAMPING, () => this.getSpringDamping(), a => this.setSpringDamping(a)));
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.SPRING_LENGTH, RollerSingleSim.i18n.SPRING_LENGTH, () => this.getSpringLength(), a => this.setSpringLength(a)));
        this.addParameter(new ParameterNumber(this, RollerSingleSim.en.SPRING_STIFFNESS, RollerSingleSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', ball1_: ' + this.ball1_
            + ', anchor_: ' + this.anchor_
            + ', spring_: ' + this.spring_
            + ', path_: ' + this.path_
            + ', hasSpring_: ' + this.hasSpring_
            + ', damping_: ' + Util.NF(this.damping_)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', lowestPoint_: ' + Util.NF(this.lowestPoint_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'RollerSingleSim';
    }
    ;
    getPath() {
        return this.path_;
    }
    ;
    setPath(path) {
        const simList = this.getSimList();
        const oldPath = this.path_;
        if (oldPath != null) {
            simList.remove(oldPath);
        }
        this.path_ = path;
        simList.add(path);
        const r = path.getBoundsWorld();
        const va = this.getVarsList();
        if (this.hasSpring_) {
            va.setValue(ANCHOR_X, r.getLeft() + r.getWidth() * 0.2);
            va.setValue(ANCHOR_Y, r.getTop() - r.getHeight() * 0.4);
        }
        this.lowestPoint_ = r.getBottom();
        const start = new Vector(r.getLeft() + r.getWidth() * 0.1, r.getTop() - r.getHeight() * 0.1);
        this.pathPoint_ = path.findNearestGlobal(start);
        va.setValue(TRACK_P, this.pathPoint_.p);
        va.setValue(TRACK_V, 0);
        va.setValue(TIME, 0);
        this.modifyObjects();
        this.saveInitialState();
    }
    ;
    modifyObjects() {
        if (this.path_ != null) {
            const va = this.getVarsList();
            const vars = va.getValues();
            this.moveObjects(vars);
            const p = this.path_.mod_p(vars[TRACK_P]);
            if (p != vars[TRACK_P]) {
                vars[TRACK_P] = p;
                va.setValue(TRACK_P, p);
            }
            va.setValue(X, this.path_.map_p_to_x(vars[TRACK_P]), true);
            va.setValue(Y, this.path_.map_p_to_y(vars[TRACK_P]), true);
            const ei = this.getEnergyInfo();
            va.setValue(KE, ei.getTranslational(), true);
            va.setValue(PE, ei.getPotential(), true);
            va.setValue(TE, ei.getTotalEnergy(), true);
        }
    }
    ;
    moveObjects(vars) {
        if (this.path_ != null) {
            this.pathPoint_.p = vars[TRACK_P];
            this.path_.map_p_to_slope(this.pathPoint_);
            this.ball1_.setPosition(this.pathPoint_);
            this.ball1_.setVelocity(this.pathPoint_.getSlope().multiply(vars[TRACK_V]));
        }
        if (this.hasSpring_) {
            this.anchor_.setPosition(new Vector(vars[ANCHOR_X], vars[ANCHOR_Y]));
        }
    }
    ;
    getEnergyInfo() {
        const ke = this.ball1_.getKineticEnergy();
        let pe = this.ball1_.getMass() * this.gravity_ *
            (this.ball1_.getPosition().getY() - this.lowestPoint_);
        if (this.hasSpring_) {
            pe += this.spring_.getPotentialEnergy();
        }
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.ball1_) {
            this.dragObj_ = simObject;
            return true;
        }
        else if (simObject == this.anchor_) {
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const p = location.subtract(offset);
        if (simObject == this.ball1_ && this.path_ != null) {
            this.pathPoint_ = this.path_.findNearestGlobal(p);
            va.setValue(TRACK_P, this.pathPoint_.p);
            va.setValue(TRACK_V, 0);
            va.incrSequence(X, Y, KE, PE, TE);
        }
        else if (simObject == this.anchor_) {
            va.setValue(ANCHOR_X, p.getX());
            va.setValue(ANCHOR_Y, p.getY());
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragObj_ = null;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (this.dragObj_ != this.ball1_) {
            this.moveObjects(vars);
            change[TRACK_P] = vars[TRACK_V];
            const k = this.pathPoint_.slope;
            const sinTheta = isFinite(k) ? k / Math.sqrt(1 + k * k) : 1;
            const mass = this.ball1_.getMass();
            change[TRACK_V] = -this.gravity_ * this.pathPoint_.direction * sinTheta
                - this.damping_ * vars[TRACK_V] / mass;
            if (this.hasSpring_) {
                let tangent;
                if (!isFinite(k)) {
                    tangent = new Vector(0, k > 0 ? 1 : -1, 0);
                }
                else {
                    tangent = new Vector(1, k, 0);
                    tangent = tangent.normalize().multiply(this.pathPoint_.direction);
                }
                const force = this.spring_.calculateForces()[0];
                Util.assert(force.getBody() == this.ball1_);
                const f = force.getVector();
                change[TRACK_V] += f.dotProduct(tangent) / mass;
            }
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerSingleSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(RollerSingleSim.en.DAMPING);
    }
    getMass() {
        return this.ball1_.getMass();
    }
    setMass(value) {
        this.ball1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(RollerSingleSim.en.MASS);
    }
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerSingleSim.en.SPRING_STIFFNESS);
    }
    getSpringLength() {
        return this.spring_.getRestLength();
    }
    setSpringLength(value) {
        this.spring_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerSingleSim.en.SPRING_LENGTH);
    }
    getSpringDamping() {
        return this.spring_.getDamping();
    }
    setSpringDamping(value) {
        this.spring_.setDamping(value);
        this.broadcastParameter(RollerSingleSim.en.SPRING_DAMPING);
    }
}
RollerSingleSim.en = {
    ANCHOR_X: 'anchor X',
    ANCHOR_Y: 'anchor Y',
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    MASS: 'mass',
    SPRING_DAMPING: 'spring damping',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    POSITION: 'position',
    VELOCITY: 'velocity',
    X_POSITION: 'x position',
    Y_POSITION: 'y position'
};
RollerSingleSim.de_strings = {
    ANCHOR_X: 'Anker X',
    ANCHOR_Y: 'Anker Y',
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    SPRING_DAMPING: 'Federdämpfung',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    POSITION: 'Position',
    VELOCITY: 'Geschwindigkeit',
    X_POSITION: 'x Position',
    Y_POSITION: 'y Position'
};
RollerSingleSim.i18n = Util.LOCALE === 'de' ? RollerSingleSim.de_strings : RollerSingleSim.en;
Util.defineGlobal('sims$roller$RollerSingleSim', RollerSingleSim);
