import { AbstractApp } from '../common/AbstractApp.js';
import { ButtonControl } from '../../lab/controls/ButtonControl.js';
import { ChainOfSpringsSim } from './ChainOfSpringsSim.js';
import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DisplaySpring } from '../../lab/view/DisplaySpring.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass, ShapeType } from '../../lab/model/PointMass.js';
import { SimList } from '../../lab/model/SimList.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
export class ChainOfSpringsApp extends AbstractApp {
    constructor(elem_ids, numAtoms, attachRight) {
        Util.setErrorHandler();
        numAtoms = typeof numAtoms === 'number' ? numAtoms : 10;
        attachRight = attachRight !== undefined ? attachRight : true;
        const simRect = new DoubleRect(-6.4, -6, 6.4, 6);
        const sim = new ChainOfSpringsSim();
        sim.makeChain(numAtoms, attachRight);
        const advance = new SimpleAdvance(sim);
        super(elem_ids, simRect, sim, advance, sim, sim);
        this.numAtoms = numAtoms;
        this.attachRight = attachRight;
        this.mySim = sim;
        this.protoMass = new DisplayShape();
        this.protoMass.setFillStyle('blue');
        this.protoAnchor = new DisplayShape();
        this.protoAnchor.setFillStyle('gray');
        this.protoSpring = new DisplaySpring();
        this.protoSpring.setWidth(0.3);
        this.protoSpring.setColorCompressed('#0c0');
        this.protoSpring.setColorExpanded('#6f6');
        this.protoSpring.setZIndex(-1);
        this.simList.toArray().forEach(obj => this.addBody(obj));
        this.simList.addObserver(this);
        this.addPlaybackControls();
        let pb;
        let pn;
        this.addParameter(pn = new ParameterNumber(this, ChainOfSpringsSim.en.NUM_LINKS, ChainOfSpringsSim.i18n.NUM_LINKS, () => this.getNumLinks(), a => this.setNumLinks(a)));
        pn.setDecimalPlaces(0);
        this.addControl(new SliderControl(pn, 0, 30, false));
        this.addParameter(pb = new ParameterBoolean(this, ChainOfSpringsSim.en.ATTACH_RIGHT, ChainOfSpringsSim.i18n.ATTACH_RIGHT, () => this.getAttachRight(), a => this.setAttachRight(a)));
        this.addControl(new CheckBoxControl(pb));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.GRAVITY);
        this.addControl(new SliderControl(pn, 0, 40, false));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.DAMPING);
        this.addControl(new SliderControl(pn, 0, 1, false));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.SPRING_DAMPING);
        this.addControl(new SliderControl(pn, 0, 1, false));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.MASS);
        this.addControl(new SliderControl(pn, 0.2, 50.2, true));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.LENGTH);
        this.addControl(new SliderControl(pn, 0, 10, false));
        pn = this.mySim.getParameterNumber(ChainOfSpringsSim.en.STIFFNESS);
        this.addControl(new SliderControl(pn, 0, 100, false));
        pn = sim.getParameterNumber(EnergyInfo.en.PE_OFFSET);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        const bc = new ButtonControl(ChainOfSpringsSim.i18n.STRAIGHT_LINE, () => this.mySim.straightLine());
        this.addControl(bc);
        this.graph.line.setXVariable(9);
        this.graph.line.setYVariable(11);
        this.graph.line.setDrawingMode("lines");
        this.timeGraph.line1.setYVariable(1);
        this.timeGraph.line2.setYVariable(2);
        this.makeEasyScript();
        this.addURLScriptButton();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', numAtoms: ' + this.numAtoms
            + ', attachRight: ' + this.attachRight
            + super.toString();
    }
    ;
    getClassName() {
        return 'ChainOfSpringsApp';
    }
    ;
    addBody(obj) {
        if (this.displayList.find(obj)) {
            return;
        }
        if (obj instanceof PointMass) {
            const proto = obj.getShape() == ShapeType.OVAL ? this.protoMass : this.protoAnchor;
            this.displayList.add(new DisplayShape(obj, proto));
        }
        else if (obj instanceof Spring) {
            this.displayList.add(new DisplaySpring(obj, this.protoSpring));
        }
    }
    ;
    observe(event) {
        if (event.getSubject() == this.simList) {
            const obj = event.getValue();
            if (event.nameEquals(SimList.OBJECT_ADDED)) {
                this.addBody(obj);
            }
            else if (event.nameEquals(SimList.OBJECT_REMOVED)) {
                const d = this.displayList.find(obj);
                if (d) {
                    this.displayList.remove(d);
                }
            }
        }
    }
    ;
    getNumLinks() {
        return this.numAtoms;
    }
    ;
    setNumLinks(value) {
        value = Math.floor(value + 0.5);
        if (this.numAtoms != value) {
            this.numAtoms = value;
            this.mySim.makeChain(this.numAtoms, this.attachRight);
            this.broadcastParameter(ChainOfSpringsSim.en.NUM_LINKS);
            this.easyScript.update();
        }
    }
    ;
    getAttachRight() {
        return this.attachRight;
    }
    ;
    setAttachRight(value) {
        if (this.attachRight != value) {
            this.attachRight = value;
            this.mySim.makeChain(this.numAtoms, this.attachRight);
            this.broadcastParameter(ChainOfSpringsSim.en.ATTACH_RIGHT);
            this.easyScript.update();
        }
    }
    ;
}
Util.defineGlobal('sims$springs$ChainOfSpringsApp', ChainOfSpringsApp);
