import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { MutableVector } from '../../lab/util/MutableVector.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TIME = 0;
const KE = 1;
const PE = 2;
const TE = 3;
const U0 = 4;
const V0 = 5;
const U1 = 6;
const V1 = 7;
const U2 = 8;
const V2 = 9;
export class CollideSpringSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.dragIdx_ = -1;
        this.dragSpring_ = null;
        this.mass_ = 0.1;
        this.damping_ = 0;
        this.restLength_ = 1.0;
        this.blockWidth_ = 0.6;
        this.stiffness_ = 60.0;
        this.springDamping_ = 0;
        this.potentialOffset_ = 0;
        this.blocks_ = [];
        this.springs_ = [];
        this.mouse_ = PointMass.makeCircle(0.5, 'mouse');
        this.mouse_.setMass(Infinity);
        this.wall1_ = PointMass.makeRectangle(0.4, 4, 'wall1');
        this.wall1_.setMass(Infinity);
        this.wall1_.setPosition(new Vector(-6.2, 0));
        this.wall2_ = PointMass.makeRectangle(0.4, 4, 'wall2');
        this.wall2_.setMass(Infinity);
        this.wall2_.setPosition(new Vector(6.2, 0));
        this.getVarsList().addVariables(this.makeVarNames(0, false), this.makeVarNames(0, true));
        let pn;
        this.addParameter(new ParameterNumber(this, CollideSpringSim.en.MASS, CollideSpringSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, CollideSpringSim.en.SPRING_STIFFNESS, CollideSpringSim.i18n.SPRING_STIFFNESS, () => this.getStiffness(), a => this.setStiffness(a)));
        this.addParameter(new ParameterNumber(this, CollideSpringSim.en.DAMPING, CollideSpringSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, CollideSpringSim.en.SPRING_DAMPING, CollideSpringSim.i18n.SPRING_DAMPING, () => this.getSpringDamping(), a => this.setSpringDamping(a)));
        this.addParameter(new ParameterNumber(this, CollideSpringSim.en.SPRING_LENGTH, CollideSpringSim.i18n.SPRING_LENGTH, () => this.getLength(), a => this.setLength(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', blocks: ' + this.blocks_.length
            + ', damping_: ' + Util.NF(this.damping_)
            + ', mass_: ' + Util.NF(this.mass_)
            + ', springDamping_: ' + Util.NF(this.springDamping_)
            + ', stiffness_: ' + Util.NF(this.stiffness_)
            + ', restLength_: ' + Util.NF(this.restLength_)
            + ', wall1_: ' + this.wall1_
            + ', wall2_: ' + this.wall2_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'CollideSpringSim';
    }
    ;
    makeVarNames(numBlocks, localized) {
        const names = [];
        const n = numBlocks * 2 + 4;
        for (let i = 0; i < n; i++) {
            names.push(this.getVariableName(i, localized));
        }
        return names;
    }
    ;
    getVariableName(idx, localized) {
        if (idx >= U0) {
            const j = (idx - U0) % 2;
            const block = 1 + Math.floor((idx - U0) / 2);
            switch (j) {
                case 0:
                    return (localized ? CollideSpringSim.i18n.POSITION :
                        CollideSpringSim.en.POSITION) + ' ' + block;
                case 1:
                    return (localized ? CollideSpringSim.i18n.VELOCITY :
                        CollideSpringSim.en.VELOCITY) + ' ' + block;
            }
        }
        else {
            switch (idx) {
                case KE:
                    return localized ? EnergyInfo.i18n.KINETIC_ENERGY :
                        EnergyInfo.en.KINETIC_ENERGY;
                case PE:
                    return localized ? EnergyInfo.i18n.POTENTIAL_ENERGY :
                        EnergyInfo.en.POTENTIAL_ENERGY;
                case TE:
                    return localized ? EnergyInfo.i18n.TOTAL_ENERGY :
                        EnergyInfo.en.TOTAL_ENERGY;
                case TIME:
                    return localized ? VarsList.i18n.TIME :
                        VarsList.en.TIME;
            }
        }
        throw '';
    }
    ;
    config(numBlocks, startPosition, startGap) {
        if (numBlocks < 1 || numBlocks > 3) {
            throw 'too many blocks ' + numBlocks;
        }
        this.getSimList().removeAll(this.blocks_);
        this.blocks_.length = 0;
        this.getSimList().removeAll(this.springs_);
        this.springs_.length = 0;
        this.getSimList().add(this.wall1_);
        this.getSimList().add(this.wall2_);
        const va = this.getVarsList();
        va.deleteVariables(0, va.numVariables());
        va.addVariables(this.makeVarNames(numBlocks, false), this.makeVarNames(numBlocks, true));
        const left = this.wall1_.getPosition();
        const right = this.wall2_.getPosition();
        for (let i = 0; i < numBlocks; i++) {
            const block = PointMass.makeSquare(this.blockWidth_, 'block_' + (i + 1));
            block.setMass(this.mass_);
            this.blocks_.push(block);
        }
        this.getSimList().addAll(this.blocks_);
        {
            const spring = new Spring('spring_0', this.wall1_, new Vector(this.wall1_.getRightBody(), 0), this.blocks_[0], new Vector(-this.blockWidth_ / 2, 0), this.restLength_, this.stiffness_, true);
            spring.setDamping(this.springDamping_);
            this.springs_.push(spring);
        }
        for (let i = 0; i < numBlocks - 1; i++) {
            const spring = new Spring('spring_' + i, this.blocks_[i], new Vector(this.blockWidth_ / 2, 0), this.blocks_[i + 1], new Vector(-this.blockWidth_ / 2, 0), this.restLength_, this.stiffness_, true);
            spring.setDamping(this.springDamping_);
            this.springs_.push(spring);
        }
        {
            const spring = new Spring('spring_' + numBlocks, this.blocks_[numBlocks - 1], new Vector(this.blockWidth_ / 2, 0), this.wall2_, new Vector(this.wall2_.getLeftBody(), 0), this.restLength_, this.stiffness_, true);
            spring.setDamping(this.springDamping_);
            this.springs_.push(spring);
        }
        this.getSimList().addAll(this.springs_);
        const vars = va.getValues();
        vars[U0] = this.wall1_.getRightWorld() + this.restLength_ + this.blockWidth_ / 2
            + startGap;
        vars[V0] = 3;
        switch (startPosition) {
            case CollideSpringSim.START_MIDDLE:
                if (numBlocks >= 2) {
                    vars[U1] = 0;
                    vars[V1] = 0;
                }
                if (numBlocks >= 3) {
                    vars[U2] = this.blockWidth_ + this.restLength_ + startGap;
                    vars[V2] = 0;
                }
                break;
            case CollideSpringSim.START_ON_WALL:
                const wallLeft = this.wall2_.getLeftWorld();
                const bw = this.blockWidth_;
                if (numBlocks == 2) {
                    vars[U1] = wallLeft - bw / 2 - this.restLength_ - startGap;
                    vars[V1] = 0;
                }
                else if (numBlocks == 3) {
                    vars[U1] = wallLeft - 3 * bw / 2 - 2 * this.restLength_ - 2 * startGap;
                    vars[V1] = 0;
                    vars[U2] = wallLeft - bw / 2 - this.restLength_ - startGap;
                    vars[V2] = 0;
                }
                break;
            default:
                throw '';
        }
        va.setValues(vars);
        this.saveInitialState();
        this.modifyObjects();
    }
    ;
    getEnergyInfo() {
        let ke = 0;
        let pe = 0;
        this.springs_.forEach(spr => pe += spr.getPotentialEnergy());
        this.blocks_.forEach(block => ke += block.getKineticEnergy());
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        this.blocks_.forEach((block, i) => {
            const idx = 2 * i;
            block.setPosition(new Vector(vars[U0 + idx], 0));
            block.setVelocity(new Vector(vars[V0 + idx], 0));
        });
    }
    ;
    startDrag(simObject, location, _offset, _dragBody, _modifiers) {
        if (simObject instanceof PointMass) {
            this.dragIdx_ = this.blocks_.indexOf(simObject);
            if (this.dragIdx_ < 0) {
                return false;
            }
            else {
                this.mouse_.setPosition(location);
                this.dragSpring_ = new Spring('drag spring', this.mouse_, Vector.ORIGIN, this.blocks_[this.dragIdx_], Vector.ORIGIN, 0, 1);
                this.getSimList().add(this.dragSpring_);
                return true;
            }
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const p = location.subtract(offset);
        if (simObject == this.wall1_) {
            this.wall1_.setPosition(p);
        }
        else if (simObject == this.wall2_) {
            this.wall2_.setPosition(p);
        }
        else if (this.dragIdx_ > -1) {
            this.mouse_.setPosition(location);
        }
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragIdx_ = -1;
        if (this.dragSpring_ != null) {
            this.getSimList().remove(this.dragSpring_);
            this.dragSpring_ = null;
        }
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        this.blocks_.forEach((block, listIdx) => {
            const idx = 2 * listIdx;
            change[U0 + idx] = vars[V0 + idx];
            const mass = block.getMass();
            const force = new MutableVector(0, 0);
            this.springs_.forEach(spr => {
                if (spr.getBody1() == block) {
                    force.add(spr.calculateForces()[0].getVector());
                }
                else if (spr.getBody2() == block) {
                    force.add(spr.calculateForces()[1].getVector());
                }
            });
            if (this.dragSpring_ != null && this.dragIdx_ == listIdx) {
                Util.assert(this.dragSpring_.getBody2() == block);
                force.add(this.dragSpring_.calculateForces()[1].getVector());
            }
            force.add(new Vector(vars[V0 + idx], 0).multiply(-this.damping_));
            change[V0 + idx] = force.getX() / mass;
        });
        return null;
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(CollideSpringSim.en.DAMPING);
    }
    ;
    getSpringDamping() {
        return this.springDamping_;
    }
    ;
    setSpringDamping(value) {
        this.springDamping_ = value;
        this.springs_.forEach(spr => spr.setDamping(value));
        this.broadcastParameter(CollideSpringSim.en.SPRING_DAMPING);
    }
    ;
    getMass() {
        return this.mass_;
    }
    ;
    setMass(value) {
        this.mass_ = value;
        const mass = this.mass_ / this.blocks_.length;
        this.blocks_.forEach(block => block.setMass(mass));
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(CollideSpringSim.en.MASS);
    }
    ;
    getLength() {
        return this.restLength_;
    }
    ;
    setLength(value) {
        this.restLength_ = value;
        for (let i = 0; i < this.springs_.length; i++) {
            this.springs_[i].setRestLength(value);
        }
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideSpringSim.en.SPRING_LENGTH);
    }
    ;
    getStiffness() {
        return this.stiffness_;
    }
    ;
    setStiffness(value) {
        this.stiffness_ = value;
        for (let i = 0; i < this.springs_.length; i++) {
            this.springs_[i].setStiffness(value);
        }
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(CollideSpringSim.en.SPRING_STIFFNESS);
    }
    ;
}
CollideSpringSim.START_MIDDLE = 0;
CollideSpringSim.START_ON_WALL = 1;
CollideSpringSim.en = {
    NUM_BLOCKS: 'number blocks',
    POSITION: 'position',
    VELOCITY: 'velocity',
    DAMPING: 'damping',
    SPRING_DAMPING: 'spring damping',
    GRAVITY: 'gravity',
    MASS: 'mass',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    START_POSITION: 'start position',
    START_IN_MIDDLE: 'in middle',
    START_ON_WALL: 'on wall',
    START_GAP: 'starting gap'
};
CollideSpringSim.de_strings = {
    NUM_BLOCKS: 'Blockanzahl',
    POSITION: 'Position',
    VELOCITY: 'Geschwindigkeit',
    DAMPING: 'Dämpfung',
    SPRING_DAMPING: 'Federdämpfung',
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    START_POSITION: 'Anfangspunkt',
    START_IN_MIDDLE: 'in der Mitte',
    START_ON_WALL: 'an der Mauer',
    START_GAP: 'Anfangsabstand'
};
CollideSpringSim.i18n = Util.LOCALE === 'de' ? CollideSpringSim.de_strings : CollideSpringSim.en;
Util.defineGlobal('sims$springs$CollideSpringSim', CollideSpringSim);
