import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const X = 0;
const V = 1;
const WORK = 2;
const TIME = 3;
const ACCEL = 4;
const KE = 5;
const PE = 6;
const TE = 7;
export class SingleSpringSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.damping_ = 0.1;
        this.initialEnergy_ = 0;
        this.potentialOffset_ = 0;
        this.isDragging = false;
        const var_names = [
            SingleSpringSim.en.POSITION,
            SingleSpringSim.en.VELOCITY,
            SingleSpringSim.en.WORK_FROM_DAMPING,
            VarsList.en.TIME,
            SingleSpringSim.en.ACCELERATION,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            SingleSpringSim.i18n.POSITION,
            SingleSpringSim.i18n.VELOCITY,
            SingleSpringSim.i18n.WORK_FROM_DAMPING,
            VarsList.i18n.TIME,
            SingleSpringSim.i18n.ACCELERATION,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(WORK, ACCEL, KE, PE, TE);
        this.block_ = PointMass.makeRectangle(0.4, 0.8, 'block');
        this.block_.setMass(0.5);
        ;
        this.fixedPoint_ = PointMass.makeSquare(0.5, 'fixed_point');
        this.fixedPoint_.setMass(Infinity);
        const restLength = 2.5;
        this.fixedPoint_.setPosition(new Vector(-restLength, 0));
        this.spring_ = new Spring('spring', this.fixedPoint_, Vector.ORIGIN, this.block_, Vector.ORIGIN, restLength, 3.0);
        this.getVarsList().setValue(X, -2.0);
        this.initWork();
        this.saveInitialState();
        this.getSimList().add(this.block_, this.fixedPoint_, this.spring_);
        let pn;
        this.addParameter(new ParameterNumber(this, SingleSpringSim.en.DAMPING, SingleSpringSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, SingleSpringSim.en.SPRING_LENGTH, SingleSpringSim.i18n.SPRING_LENGTH, () => this.getSpringRestLength(), a => this.setSpringRestLength(a)));
        this.addParameter(new ParameterNumber(this, SingleSpringSim.en.MASS, SingleSpringSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, SingleSpringSim.en.SPRING_STIFFNESS, SingleSpringSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, SingleSpringSim.en.FIXED_POINT, SingleSpringSim.i18n.FIXED_POINT, () => this.getFixedPoint(), a => this.setFixedPoint(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', block_: ' + this.block_
            + ', fixedPoint_: ' + this.fixedPoint_
            + ', spring_: ' + this.spring_
            + ', damping_: ' + Util.NF(this.damping_)
            + ', initialEnergy_: ' + Util.NF(this.initialEnergy_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'SingleSpringSim';
    }
    ;
    reset() {
        super.reset();
        this.initWork();
    }
    ;
    initWork() {
        this.getVarsList().setValue(WORK, 0);
        this.initialEnergy_ = this.getEnergyInfo().getTotalEnergy();
    }
    ;
    getEnergyInfo() {
        const vars = this.getVarsList().getValues();
        return this.getEnergyInfo_(vars[WORK]);
    }
    ;
    getEnergyInfo_(work) {
        const ke = this.block_.getKineticEnergy();
        const pe = this.spring_.getPotentialEnergy() + this.potentialOffset_;
        return new EnergyInfo(pe, ke, NaN, work, this.initialEnergy_);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const rate = new Array(vars.length);
        this.evaluate(vars, rate, 0);
        vars[ACCEL] = rate[V];
        const ei = this.getEnergyInfo_(vars[WORK]);
        vars[KE] = ei.getTranslational();
        vars[PE] = ei.getPotential();
        vars[TE] = ei.getTotalEnergy();
        va.setValues(vars, true);
    }
    ;
    moveObjects(vars) {
        this.block_.setPosition(new Vector(vars[X], 0));
        this.block_.setVelocity(new Vector(vars[V], 0, 0));
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.block_) {
            this.isDragging = true;
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        if (simObject == this.block_) {
            const p = location.subtract(offset);
            va.setValue(X, p.getX());
            va.setValue(V, 0);
            this.initWork();
            va.incrSequence(ACCEL, KE, PE, TE);
            this.moveObjects(va.getValues());
        }
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1.0;
        if (!this.isDragging) {
            this.moveObjects(vars);
            change[X] = vars[V];
            change[V] = (-this.spring_.getStiffness() * this.spring_.getStretch() -
                this.damping_ * vars[V]) / this.block_.getMass();
            change[WORK] = -this.damping_ * vars[V] * vars[V];
        }
        return null;
    }
    ;
    getMass() {
        return this.block_.getMass();
    }
    ;
    setMass(value) {
        this.block_.setMass(value);
        this.initWork();
        this.getVarsList().incrSequence(ACCEL, KE, PE, TE);
        this.broadcastParameter(SingleSpringSim.en.MASS);
    }
    ;
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    ;
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.initWork();
        this.getVarsList().incrSequence(ACCEL, PE, TE);
        this.broadcastParameter(SingleSpringSim.en.SPRING_STIFFNESS);
    }
    ;
    getSpringRestLength() {
        return this.spring_.getRestLength();
    }
    ;
    setSpringRestLength(value) {
        this.spring_.setRestLength(value);
        this.initWork();
        this.getVarsList().incrSequence(ACCEL, PE, TE);
        this.broadcastParameter(SingleSpringSim.en.SPRING_LENGTH);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.initWork();
        this.broadcastParameter(SingleSpringSim.en.DAMPING);
    }
    ;
    getFixedPoint() {
        return this.fixedPoint_.getPosition().getX();
    }
    ;
    setFixedPoint(value) {
        this.fixedPoint_.setPosition(new Vector(value, 0));
        this.initWork();
        this.broadcastParameter(SingleSpringSim.en.FIXED_POINT);
    }
    ;
}
SingleSpringSim.en = {
    ACCELERATION: 'acceleration',
    DAMPING: 'damping',
    MASS: 'mass',
    POSITION: 'position',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    VELOCITY: 'velocity',
    WORK_FROM_DAMPING: 'work from damping',
    FIXED_POINT: 'fixed point'
};
SingleSpringSim.de_strings = {
    ACCELERATION: 'Beschleunigung',
    DAMPING: 'Dämpfung',
    MASS: 'Masse',
    POSITION: 'Position',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    VELOCITY: 'Geschwindigkeit',
    WORK_FROM_DAMPING: 'Arbeit der Dämpfung',
    FIXED_POINT: 'Festpunkt'
};
SingleSpringSim.i18n = Util.LOCALE === 'de' ? SingleSpringSim.de_strings : SingleSpringSim.en;
Util.defineGlobal('sims$springs$SingleSpringSim', SingleSpringSim);
