import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const UX = 0;
const UY = 1;
const VX = 2;
const VY = 3;
const KE = 4;
const PE = 5;
const TE = 6;
const TIME = 7;
const ANCHORX = 8;
const ANCHORY = 9;
export class Spring2DSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.isDragging_ = false;
        this.gravity_ = 9.8;
        this.damping_ = 0;
        this.potentialOffset_ = 0;
        this.anchor_ = PointMass.makeSquare(0.5, 'anchor');
        this.bob_ = PointMass.makeCircle(0.5, 'bob');
        const var_names = [
            Spring2DSim.en.X_POSITION,
            Spring2DSim.en.Y_POSITION,
            Spring2DSim.en.X_VELOCITY,
            Spring2DSim.en.Y_VELOCITY,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME,
            Spring2DSim.en.ANCHOR_X,
            Spring2DSim.en.ANCHOR_Y
        ];
        const i18n_names = [
            Spring2DSim.i18n.X_POSITION,
            Spring2DSim.i18n.Y_POSITION,
            Spring2DSim.i18n.X_VELOCITY,
            Spring2DSim.i18n.Y_VELOCITY,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME,
            Spring2DSim.i18n.ANCHOR_X,
            Spring2DSim.i18n.ANCHOR_Y
        ];
        const va = new VarsList(var_names, i18n_names, this.getName() + '_VARS');
        this.setVarsList(va);
        va.setComputed(KE, PE, TE);
        this.anchor_.setPosition(new Vector(0, 3));
        this.bob_.setMass(0.5);
        this.spring_ = new Spring('spring', this.anchor_, Vector.ORIGIN, this.bob_, Vector.ORIGIN, 2.5, 6.0);
        this.getSimList().add(this.anchor_, this.bob_, this.spring_);
        let pn;
        this.addParameter(new ParameterNumber(this, Spring2DSim.en.GRAVITY, Spring2DSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, Spring2DSim.en.MASS, Spring2DSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, Spring2DSim.en.DAMPING, Spring2DSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, Spring2DSim.en.SPRING_LENGTH, Spring2DSim.i18n.SPRING_LENGTH, () => this.getSpringRestLength(), a => this.setSpringRestLength(a)));
        this.addParameter(new ParameterNumber(this, Spring2DSim.en.SPRING_STIFFNESS, Spring2DSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
        this.restState();
        va.setValue(VX, 1.5);
        va.setValue(VY, 1.7);
        this.saveInitialState();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', spring_: ' + this.spring_
            + ', bob_: ' + this.bob_
            + ', anchor_: ' + this.anchor_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'Spring2DSim';
    }
    ;
    restState() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const fixY = vars[ANCHORY] = this.anchor_.getPosition().getY();
        vars[UX] = vars[ANCHORX] = this.anchor_.getPosition().getX();
        vars[UY] = fixY - this.spring_.getRestLength()
            - this.bob_.getMass() * this.gravity_ / this.spring_.getStiffness();
        vars[VX] = vars[VY] = 0;
        va.setValues(vars);
        this.modifyObjects();
        this.potentialOffset_ = 0;
        this.setPEOffset(-this.getEnergyInfo().getPotential());
    }
    ;
    getEnergyInfo() {
        const ke = this.bob_.getKineticEnergy();
        const y = this.bob_.getPosition().getY();
        let pe = this.gravity_ * this.bob_.getMass() * y;
        pe += this.spring_.getPotentialEnergy();
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        this.bob_.setPosition(new Vector(vars[UX], vars[UY]));
        this.bob_.setVelocity(new Vector(vars[VX], vars[VY], 0));
        this.anchor_.setPosition(new Vector(vars[ANCHORX], vars[ANCHORY]));
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.bob_) {
            this.isDragging_ = true;
            return true;
        }
        else if (simObject == this.anchor_) {
            return true;
        }
        return false;
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const p = location.subtract(offset);
        if (simObject == this.anchor_) {
            va.setValue(ANCHORX, p.getX());
            va.setValue(ANCHORY, p.getY());
        }
        else if (simObject == this.bob_) {
            va.setValue(UX, p.getX());
            va.setValue(UY, p.getY());
            va.setValue(VX, 0);
            va.setValue(VY, 0);
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging_ = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        if (!this.isDragging_) {
            const forces = this.spring_.calculateForces();
            const f = forces[1];
            Util.assert(f.getBody() == this.bob_);
            const m = this.bob_.getMass();
            change[UX] = vars[VX];
            change[UY] = vars[VY];
            change[VX] = (f.getVector().getX() - this.damping_ * vars[VX]) / m;
            change[VY] = -this.gravity_ + (f.getVector().getY() - this.damping_ * vars[VY]) / m;
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Spring2DSim.en.GRAVITY);
    }
    ;
    getMass() {
        return this.bob_.getMass();
    }
    ;
    setMass(value) {
        this.bob_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(Spring2DSim.en.MASS);
    }
    ;
    getSpringRestLength() {
        return this.spring_.getRestLength();
    }
    ;
    setSpringRestLength(value) {
        this.spring_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Spring2DSim.en.SPRING_LENGTH);
    }
    ;
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    ;
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(Spring2DSim.en.SPRING_STIFFNESS);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(Spring2DSim.en.DAMPING);
    }
    ;
}
Spring2DSim.en = {
    ANCHOR_X: 'anchor X',
    ANCHOR_Y: 'anchor Y',
    X_POSITION: 'X position',
    Y_POSITION: 'Y position',
    X_VELOCITY: 'X velocity',
    Y_VELOCITY: 'Y velocity',
    DAMPING: 'damping',
    GRAVITY: 'gravity',
    MASS: 'mass',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    REST_STATE: 'rest state'
};
Spring2DSim.de_strings = {
    ANCHOR_X: 'Anker X',
    ANCHOR_Y: 'Anker Y',
    X_POSITION: 'X Position',
    Y_POSITION: 'Y Position',
    X_VELOCITY: 'X Geschwindigkeit',
    Y_VELOCITY: 'Y Geschwindigkeit',
    DAMPING: 'Dämpfung',
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    REST_STATE: 'ruhe Zustand'
};
Spring2DSim.i18n = Util.LOCALE === 'de' ? Spring2DSim.de_strings : Spring2DSim.en;
Util.defineGlobal('sims$springs$Spring2DSim', Spring2DSim);
