import { ContactSim } from '../lab/engine2D/ContactSim.js';
import { reportTestResults, myPrintln } from './TestRig.js';
import { Util } from '../lab/util/Util.js';
const debug = false;
const PRINT_ALL_VARS = false;
export function makeVars(n) {
    const vars = new Array(n);
    for (let i = 0; i < n; i++)
        vars[i] = NaN;
    return vars;
}
;
export function setBodyVars(sim, vars, i, x, vx, y, vy, w, vw) {
    const idx = sim.getBody(i).getVarsIndex();
    vars[idx + 0] = x;
    vars[idx + 1] = vx;
    vars[idx + 2] = y;
    vars[idx + 3] = vy;
    vars[idx + 4] = w;
    vars[idx + 5] = vw;
}
;
export function checkResult(sim, expected, tolerance) {
    if (expected == null || isNaN(tolerance))
        throw '';
    let passed = true;
    let idx = -1;
    let maxDiff = 0;
    const vars = sim.getVarsList().getValues(true);
    Util.assert(vars.length >= expected.length);
    for (let i = 0; i < expected.length; i++) {
        if (isNaN(expected[i]))
            continue;
        const diff = Math.abs(vars[i] - expected[i]);
        if (diff > tolerance) {
            passed = false;
            if (idx < 0 || diff > maxDiff) {
                idx = i;
                maxDiff = diff;
            }
        }
    }
    if (!passed) {
        printVars(sim);
        const s = 'vars[' + idx + ']=' + vars[idx] + ' != ' + expected[idx]
            + ' with tolerance=' + tolerance
            + ' diff=' + Util.NF5E(maxDiff);
        reportTestResults(false, 'vars', s);
    }
    return passed;
}
;
export function checkValue(message, value, expected, tolerance) {
    if (Math.abs(expected - value) > tolerance) {
        const s = message + ' expected=' + expected + ' actual=' + value
            + ' tolerance=' + tolerance;
        reportTestResults(false, 'value', s);
    }
}
;
export function runTest(sim, advance, runUntil, expectedVars, tolerance, expectedEnergyDiff, energyTol, expectedCollisions, expectedSearches) {
    if (sim instanceof ContactSim) {
        sim.setExtraAccelTimeStep(advance.getTimeStep());
    }
    expectedVars = expectedVars ?? null;
    energyTol = energyTol ?? 0;
    expectedCollisions = expectedCollisions ?? -1;
    expectedSearches = expectedSearches ?? 0;
    expectedEnergyDiff = expectedEnergyDiff ?? NaN;
    if (Util.DEBUG && debug) {
        console.log('Engine2DTestRig.runTest expectedEnergyDiff=' + Util.NFE(expectedEnergyDiff)
            + ' energyTol=' + Util.NFE(energyTol)
            + ' expectedCollisions=' + expectedCollisions);
        myPrintln(sim.toString());
    }
    if (advance.getTime() < 1E-12) {
        advance.advance(0);
        advance.getCollisionTotals().reset();
    }
    const e1 = sim.getEnergyInfo().getTotalEnergy();
    const nc1 = advance.getCollisionTotals().getCollisions();
    const bs1 = advance.getCollisionTotals().getSearches();
    if (debug && !PRINT_ALL_VARS) {
        myPrintln(' time      X        VX        Y       VY       W      VW     energy');
    }
    let lastReportTime = advance.getTime();
    while (advance.getTime() < runUntil - 0.0000001) {
        if (debug) {
            if (PRINT_ALL_VARS) {
                console.log(sim.formatVars());
            }
            else {
                printRigidBody(sim, 0);
            }
        }
        advance.advance(advance.getTimeStep());
    }
    Util.assert(Math.abs(advance.getTime() - runUntil) < 0.001);
    let didTest = false;
    let passed = true;
    let testType = '';
    if (expectedVars != null && tolerance !== undefined) {
        passed = checkResult(sim, expectedVars, tolerance);
        testType += (testType != '' ? '+' : '') + 'vars';
        didTest = true;
    }
    if (expectedEnergyDiff !== undefined && !isNaN(expectedEnergyDiff)) {
        testType += (testType != '' ? '+' : '') + 'energy';
        const e2 = sim.getEnergyInfo().getTotalEnergy();
        const energyEqual = Math.abs(e2 - e1 - expectedEnergyDiff) < energyTol;
        if (!energyEqual) {
            const s = 'energy diff=' + Util.NF9(e2 - e1)
                + ', expected diff=' + Util.NF9(expectedEnergyDiff)
                + ', energyTol=' + energyTol
                + ', error=' + Util.NF9(e2 - e1 - expectedEnergyDiff)
                + ', energy=' + Util.NF9(e2);
            reportTestResults(false, testType, s);
            passed = false;
        }
        didTest = true;
    }
    if (expectedCollisions > -1) {
        testType += (testType != '' ? '+' : '') + 'num_collisions';
        const nc2 = advance.getCollisionTotals().getCollisions();
        if (nc2 - nc1 != expectedCollisions) {
            const s = 'should have had ' + expectedCollisions + ' but had ' + (nc2 - nc1)
                + ' collisions';
            reportTestResults(false, testType, s);
            passed = false;
        }
        didTest = true;
    }
    if (expectedSearches > -1) {
        testType += (testType != '' ? '+' : '') + 'collision_search';
        const bs2 = advance.getCollisionTotals().getSearches();
        if (bs2 - bs1 != expectedSearches) {
            const s = 'should have had ' + expectedSearches + ' but had ' + (bs2 - bs1)
                + ' collision searches';
            reportTestResults(false, testType, s);
            passed = false;
        }
        didTest = true;
    }
    if (passed && didTest) {
        reportTestResults(true, testType);
    }
}
;
export function runExceptionTest(advance, time) {
    const testType = 'exception';
    while (advance.getTime() < (time - 0.0000001)) {
        try {
            advance.advance(advance.getTimeStep());
        }
        catch (e) {
            reportTestResults(true, testType);
            return;
        }
    }
    reportTestResults(false, testType, 'expected exception did not occur');
}
;
export function checkTightJoints(sim, tolerance) {
    const joints = sim.getConnectors();
    const len = joints.length;
    if (len > 0) {
        for (let i = 0; i < len; i++) {
            const joint = joints[i];
            const dist = joint.getNormalDistance();
            if (Math.abs(dist) > tolerance) {
                reportTestResults(false, 'joints', 'joint not tight, tolerance='
                    + Util.NF9(tolerance) + ' dist=' + Util.NF9(dist) + ' joint=' + joint);
                return;
            }
        }
        reportTestResults(true, 'joint_tightness');
    }
}
;
export function checkContactDistances(sim, tolerance) {
    const contacts = [];
    sim.findCollisions(contacts, sim.getVarsList().getValues(), 0);
    const len = contacts.length;
    if (len > 0) {
        for (let i = 0; i < len; i++) {
            const c = contacts[i];
            const d = Math.abs(c.distanceToHalfGap());
            const isClose = d < tolerance;
            if (!isClose) {
                reportTestResults(false, 'contact dist', 'contact is not close, distanceToHalfGap=' + Util.NFE(d)
                    + ' tolerance=' + Util.NFE(tolerance) + ' contact=' + c);
                return;
            }
        }
        reportTestResults(true, 'contact_distance');
    }
}
;
function printVars(sim) {
    const X = 0, VX = 1, Y = 2, VY = 3, W = 4, VW = 5;
    const vars = sim.getVarsList().getValues(true);
    const numBods = sim.getBodies().length;
    for (let i = 0; i < numBods; i++) {
        const idx = sim.getBody(i).getVarsIndex();
        myPrintln('Engine2DTestRig.setBodyVars(sim, vars, ' + i + ', '
            + Util.nf7(vars[idx + X]) + ', '
            + Util.nf7(vars[idx + VX]) + ', '
            + Util.nf7(vars[idx + Y]) + ', '
            + Util.nf7(vars[idx + VY]) + ', '
            + Util.nf7(vars[idx + W]) + ', '
            + Util.nf7(vars[idx + VW]) + ');');
    }
}
;
function printRigidBody(sim, index) {
    const vars = sim.getVarsList().getValues(true);
    const offset = index * 6;
    const X = 0, VX = 1, Y = 2, VY = 3, W = 4, VW = 5;
    myPrintln(Util.NF5(sim.getTime()) + ' '
        + Util.NF5(vars[offset + X]) + ' '
        + Util.NF5(vars[offset + VX]) + ' '
        + Util.NF5(vars[offset + Y]) + ' '
        + Util.NF5(vars[offset + VY]) + ' '
        + Util.NF5(vars[offset + W]) + ' '
        + Util.NF5(vars[offset + VW]) + ' '
        + Util.NF5(sim.getEnergyInfo().getTotalEnergy()));
}
;
